import enum
from xes_neo.helper import *
from xes_neo.import_lib import *
from xes_neo.ini_parser import *
#from pathObj import OliverPharr
#from .individual import Individual
#from pathrange import Pathrange_limits  #may be deletable/ built for XAFS
#from xes_neo_data import XES_Data #Do we need this still?

import numpy as np
from xes_neo.xes_individual import Individual
from xes_neo.xes_fit import peak,background
from xes_neo.xes_data import xes_data

from xes_neo.periodic_table import ElementData

from copy import deepcopy #fixes bug at line 70ish with deepcopy
"""
Author: Alana Humiston, Miu Lun Lau
"""


class XES_GA:

    def initialize_params(self,verbose = False):
        """
        Initialize Parameters
        """
        # print("Initialize Parameters")
        print("Initializing Params")
        self.intervalK = 0.05
        self.numGenSinceImproved = 0
        self.tol = np.finfo(np.float64).resolution
        

    def initialize_variable(self):
        """
        Initalize variables
        """
        print("Initializing Variables")
        self.genNum = 0
        self.nChild = 4
        self.globBestFit = [0,0]
        self.currBestFit = [0,0]
        self.bestDiff = 9999e11
        self.bestBest = 999999999e11
        self.diffCounter = 0
        self.element = element
        self.photoelectronLine = photoelectronLine
        self.transitionLine = transitionLine
        self.pathDictionary = {}
        self.data_file = data_file
        #self.data_cutoff = data_cutoff
        # Paths
        self.npaths = npaths
        #self.fits = fits

        # Populations
        self.npops = size_population
        self.ngen = number_of_generation
        self.steady_state = steady_state

        # Mutation Parameters
        self.mut_opt = mutated_options
        self.mut_chance = chance_of_mutation
        # self.mut_chance_e0 = chance_of_mutation_e0

        # Crosover Parameters
        self.n_bestsam = int(best_sample*self.npops*(0.01))
        self.n_lucksam = int(lucky_few*self.npops*(0.01))

        # Time related
        self.time = False
        self.tt = 0

    def initialize_file_path(self,i=0):
        """
        Initalize file paths for each of the file first
        """
        print("initializing file paths")
        self.base = os.getcwd()
        self.output_path = os.path.join(self.base,output_file)
        self.check_output_file(self.output_path)
        self.log_path = os.path.splitext(copy.deepcopy(self.output_path))[0] + ".log"
        self.check_if_exists(self.log_path)

        # Initialize logger
        self.logger = logging.getLogger('')
        # Delete handler
        self.logger.handlers=[]
        file_handler = logging.FileHandler(self.log_path,mode='a+',encoding='utf-8')
        stdout_handler = logging.StreamHandler(sys.stdout)

        formatter = logging.Formatter('%(message)s')
        file_handler.setFormatter(formatter)
        stdout_handler.setFormatter(formatter)
        self.logger.addHandler(file_handler)
        self.logger.addHandler(stdout_handler)

        self.logger.setLevel(logging.INFO)
        self.logger.info(banner())

    def check_if_exists(self,path_file):
        """
        Check if the directory exists
        """
        if os.path.exists(path_file):
            os.remove(path_file)
        # Make Directory when its missing
        path = pathlib.Path(path_file)
        path.parent.mkdir(parents=True, exist_ok=True)

    def check_output_file(self,file):
        """
        check if the output file for each of the file
        """
        file_base= os.path.splitext(file)[0]
        self.check_if_exists(file)
        self.file = file

        self.file_initial = open(self.output_path,"a+")
        self.file_initial.write("Gen,TPS,FITTNESS,CURRFIT,CURRIND,BESTFIT,BESTIND\n")  # writing header
        self.file_initial.close()

        file_data = os.path.splitext(file)[0] + '_data.csv'
        self.check_if_exists(file_data)
        self.file_data = file_data


    def initialize_range(self,scale_var,data_KE, data_XES,i=0,BestIndi=None):
        
        print("initializing ranges")
        """
        Initalize range

        To Do list:
            Initalize range will be difference for each paths depend if the run are
            in series, therefore the ranges will self-adjust
        """

        #self.periodicTable = ElementData(self.element,self.photoelectronLine, self.transitionLine)
        #PE_lit, singlet, so_split, Br = self.periodicTable.getParams(self.element,self.photoelectronLine) #Calls user selected element and gives elemental information
        #Current issue is that these values are given in variable form but the ones we currently call are all appended to an array.
        #This assumes that when you select something such as Zr3d that you are automatically making all other inputted peaks doublets.
        #This would need to change if you also have satellites/plasmons etc.
        #Next step is figuring out how to use this information in the algorithm.
        #We want to get rid of old user input for PE and instead go with the element
        #Need to also figure out how the user can provide elemental composition?
        #Maybe we call this in intparser instead?


        # data = np.genfromtxt(self.data_file,delimiter=',',skip_header=1)
  
        photoline_select = self.photoelectronLine
        transitionLine_select = self.transitionLine
        scale_var = scale_data
        self.data_obj = xes_data(self.data_file,skipLn = skipLn)
        #HOW TO CALL SCALE_VAR AND GET CORRECT VALUE IN THIS FUNCTION OR INIT?
        self.x_slice = self.data_obj.get_x()
       
        self.y_slice = self.data_obj.get_y(scale_var)
        self.x_array = self.x_slice
        self.y_array = self.y_slice
       

        yAvg = 0
        yTot = 0
        j=0
        for i,yVal in enumerate(self.data_obj.get_y(scale_var)[-10:]): #[-10:] gets last 10 items in array.
            yTot += yVal
            j=i
        yAvg = yTot/(j+1)

        #baseline_range[1] = yAvg
        #[0] = min range, [1] = max range, [2] = delta range
        #baseline_range = [0,yAvg,1]
        #amp_range[1] = max(self.data_obj.get_y()) #From data tesing it seems that having this in actually helps a lot with finding parameters
        
        N = 10
       
        y_left = self.y_array[:N]
        y_right = self.y_array[-N:]
        x_left = self.x_array[:N]
        x_right = self.x_array[-N:]
        self.y_left_avg = sum(y_left)/N
        x_left_avg = sum(x_left)/N
        self.y_right_avg = sum(y_right)/N
        x_right_avg = sum(x_right)/N
        
       
        #Different background ranges depending on if background is linear or not. If not it has a more limited range (closer to 0)
        
        if "Linear" in background_type:
                
          
            if self.y_left_avg > self.y_right_avg:
                if self.y_left_avg or self.y_right_avg < 0:
                   
                    baseline_range = [self.y_right_avg+np.round(self.y_right_avg*0.5, 2),self.y_left_avg-np.round(self.y_left_avg*0.25, 2),1]
                else:
                   
                    baseline_range = [self.y_right_avg-np.round(self.y_right_avg*0.5, 2),self.y_left_avg+np.round(self.y_left_avg*0.25, 2),1]
            else:
                if self.y_left_avg or self.y_right_avg < 0:
                    
                    baseline_range = [self.y_left_avg+np.round(self.y_left_avg*0.5, 2),self.y_right_avg-np.round(self.y_right_avg*0.25, 2),1]
                else:
                   
                    baseline_range = [self.y_left_avg-np.round(self.y_left_avg*0.5, 2),self.y_right_avg+np.round(self.y_right_avg*0.25, 2),1]
            if scale_var == True:
               
                baseline_range = [-0.1,0.1,0.1]
                
        else:
           
            if self.y_left_avg > self.y_right_avg:
                if self.y_right_avg < 0:
                  
                    baseline_range = [self.y_right_avg+np.round(self.y_right_avg*0.05, 2),self.y_left_avg-np.round(self.y_left_avg*0.05, 2),1]
                else:
                   
                    baseline_range = [self.y_right_avg-np.round(self.y_right_avg*0.05, 2),self.y_left_avg+np.round(self.y_left_avg*0.05, 2),1]
            else:
                if self.y_left_avg < 0:
                   
                    baseline_range = [self.y_left_avg+np.round(self.y_left_avg*0.05, 2),self.y_right_avg-np.round(self.y_right_avg*0.05, 2),1]
                else:
                  
                    baseline_range = [self.y_left_avg-np.round(self.y_left_avg*0.05, 2),self.y_right_avg+np.round(self.y_right_avg*0.05, 2),1]
            
            if scale_var == True:
             
                baseline_range = [-0.1,0.1,0.1]
   
                
        #print("BASELINE", baseline_range)
        #print("LEFT", self.y_left_avg, "RIGHT", self.y_right_avg)
        if x_right_avg > x_left_avg: #KE condition 
            y_slope = (self.y_right_avg - self.y_left_avg)/abs(x_right_avg - x_left_avg)
        else:
            y_slope = (self.y_left_avg - self.y_right_avg)/abs(x_left_avg - x_right_avg)
       
        if y_slope < 0:
           
            slope_range[0] = y_slope*1.1
            slope_range[1] = y_slope*0.9
            background_range[0] = -x_left_avg
            background_range[1] = -x_right_avg
           

        else:
            slope_range[0] = y_slope*0.9
            slope_range[1] = y_slope*1.1 #Set to the slope of the average of both sides of the data plus a little more for wiggle room
            if x_left_avg > x_right_avg:
                background_range[1] = x_left_avg
                background_range[0] = x_right_avg
            else:
                background_range[0] = x_left_avg
                background_range[1] = x_right_avg
      
       
        diff = self.y_left_avg - self.y_right_avg
        #background_shir_range[1] = self.y_right_avg + (self.y_right_avg/100)*25


        #Can change ranges to be within a certain percent range of what the user inputs. Right now it is set at 35% but can increase if needed

        '''
        percent_guess = 35/100

        sigma_range[0] =  -sigma_guess[0]*percent_guess
        sigma_range[1] =  sigma_guess[0]*percent_guess

        fwhm_range[0] = -gamma_guess[0]*percent_guess
        fwhm_range[1] =  gamma_guess[0]*percent_guess

        amp_range[0] =  -self.y_right_avg - amp_guess[0]*percent_guess #subtracts self.y_right_avg because amp guess excludes background (assuming user wont do that correctly)
        amp_range[1] =  -self.y_right_avg + amp_guess[0]*percent_guess
        '''

        #print(amp_guess[0])

        for n in range(len(amp_guess)):
            if amp_guess[n] < 0.5*round(self.y_right_avg,2):
                pass
            else:
                amp_guess[n] = amp_guess[n] -round(self.y_right_avg,2) #It is assumed that the user is inputting the amp + baseline so the average of the right data is subtratced to account for this
            
            if amp_guess[n] < 0: #No negative peaks
                amp_guess[n] = 0


        decay_rate = self.y_array[:1]*0.367879441

        for i in np.arange(0, len(self.y_array)): #Finding x value at the decay_rate intensity
            if self.y_array[i] >= decay_rate:
                decay_rate_x = self.x_array[i]
                pass
            else:
                decay_rate_x = self.x_array[i]
        amp_range = []
        PE_range = []
        sigma_range = []
        gamma_range = []

        max_peak = amp_guess[0] #Finding the peak with the largest amplitude and giving it a larger PE range
        #In the future we will need to have the algorithm check the elemental peak assingment here and give it a range based on a database of collected peak information
        peak_location = 0
        for i in amp_guess:
            if i >= max_peak:
                max_peak_location = peak_location
            peak_location += 1
        #PE_range_min[max_peak_location] = -0.5
        #PE_range_max[max_peak_location] = +0.5

        for i in range(len(PE_guess)): #Creating 2D arrays of range values for each peaks PE and amplitude

            if amp_limited[i] == True:
                amp_range.append([-0.00001, 0.00001, 0.00001]) #Making the range basically zero if the parameter is 'fixed'
            else:
                amp_range.append([amp_range_min[i], amp_range_max[i], amp_range_delta[i]])



            if PE_limited[i] == True:
                PE_range.append([-0.00001, 0.00001, 0.00001]) #Making the range basically zero if the parameter is 'fixed'
            else:
                PE_range.append([PE_range_min[i], PE_range_max[i], PE_range_delta[i]])



            if sigma_limited[i] == True:
                sigma_range.append([-0.00001, 0.00001, 0.00001]) #Making the range basically zero if the parameter is 'fixed'
            else:
                sigma_range.append([sigma_range_min[i], sigma_range_max[i], sigma_range_delta[i]])



            if gamma_limited[i] == True:
                gamma_range.append([-0.00001, 0.00001, 0.00001]) #Making the range basically zero if the parameter is 'fixed'
            else:
                gamma_range.append([gamma_range_min[i], gamma_range_max[i], gamma_range_delta[i]])


    
       
        #For data that comes in KE the spinOrbitSplitting needs to be negative
        
        for i in range(len(is_singlet)):
            if self.x_array[0] < self.x_array[-1]:
                if spinOrbitSplit_guess[i] > 0:
                    spinOrbitSplit_guess[i] = -spinOrbitSplit_guess[i]



      

        #exp_decay_range[0] = decay_rate_x - decay_rate_x*0.1
        #exp_decay_range[1] = decay_rate_x + decay_rate_x*0.1
        #background_shir_range[0] = y_right_avg - (y_right_avg/100)*25 #I think with the use of multiple peaks this should be allowed to be zero
        #k_range[1] = self.y_left_avg - y_right_avg + 200
       
        self.pars_range = {
            'Photon Energy': PE_range,
            'PE':PE_guess,
            'PE_limited' :PE_limited,
            'PE_correlated' :PE_correlated,
            'PE_correlated_mult' :PE_correlated_mult,
            'Gaussian': sigma_range,
            'Sigma': sigma_guess,
            'sigma_limited' :sigma_limited,
            'sigma_correlated' :sigma_correlated,
            'sigma_correlated_mult' :sigma_correlated_mult,
            'Lorentzian': gamma_range,
            'Gamma': gamma_guess,
            'gamma_limited' :gamma_limited,
            'gamma_correlated' :gamma_correlated,
            'gamma_correlated_mult' :gamma_correlated_mult,
            'Amplitude' : amp_range,
            'Amp': amp_guess,
            'amp_limited' :amp_limited,
            'amp_correlated' :amp_correlated,
            'amp_correlated_mult' :amp_correlated_mult,
            'Asymmetry' : asymmetry_range,
            'Asymmetry Doniach-Sunjic' : asymmetryDoniach_range,
            'k_range' : k_range,
            'Background' : background_range,
            #'Shirley Background' : background_shir_range,
            'CTou3' : CTou3_range,
            'DTou3' : DTou3_range,
            'Slope' : slope_range,
            'Exponential Amplitude' : exp_amp_range,
            'Decay' : exp_decay_range,
            'npeaks' : npaths,
            'baseline' : baseline_range,
            'branching_ratios' : branching_ratio,
            'is_singlet' : is_singlet,
            'is_coster_kronig' : is_coster_kronig,
            'Lorentzian Coster-Kronig' : gamma_CK_range,
            'Gamma Coster-Kronig' : gamma_CK_guess,
            'spinOrbitSplitting' : spinOrbitSplit_range,
            'spinOrbitSplit' : spinOrbitSplit_guess,
            'photoline' : photoline_select,
            'transitionline' :transitionLine_select
        }
        self.peak_type = peak_type
        self.backgrounds = background_type
       


    ''' Dont' think it's neccesary, we'll see
    def create_range(self,value,percentage,dt,prec): #-------Where is this called?--------
        """
        Create delta to calculate the ranges
        """
        minus = round(value - percentage*value,prec)
        plus = round(value + percentage*value,prec)
        range = np.arange(minus,plus+dt,dt)
        return range
    '''
   

    def scanResidual(self): #Need to expand function to see if overfitting the data --> Reomve that peak
        #Function to analyze the residual to determine if we need to add/subtract a peak
        
        
        
      
        

        yTotal = np.zeros(len(self.x_array)) 
        self.residual = np.zeros(len(self.x_array))
        residual_noiseless = np.zeros(len(self.x_array))

        yTotal = self.currBestY #Using the best fit Y values of the data 
       
        

        for j in range(len(self.x_array)):
            self.residual[j] = ((self.y_array[j] -yTotal[j])) #Positive means under fitting here and negative means overfitting
       

        #Want to subtract off the avg of the residual from both ends of the data to account for noise in data 
        N = 10
        res_left = self.residual[:N]
        res_right = self.residual[-N:]
       

        res_left_avg = sum(res_left)/N
       
        res_right_avg = sum(res_right)/N

        residual_average = (res_left_avg + res_right_avg)/2
       

        for j in range(2, len(self.x_array-2)):
            residual_noiseless[j] = self.residual[j] - residual_average

        max_res_location = 0
        #Finding location of top/bottom 6 maximum/minimum residual (where we want to add/remove a new peak)
        sorting_res = np.argsort(residual_noiseless)
        sorted_res = residual_noiseless[sorting_res]
       
        top_6_rev = sorted_res[-6 : ] 
        bottom_6 = sorted_res[ : 6] 
        top_6_PE = []
        bottom_6_PE = []

        top_6 = top_6_rev[::-1]
        for i in range(len(self.x_array)):
            if residual_noiseless[i] == top_6[0]:
                max_PE_1st_res = self.x_array[i]
            elif residual_noiseless[i] == top_6[1]:
                max_PE_2nd_res = self.x_array[i]
            elif residual_noiseless[i] == top_6[2]:
                max_PE_3rd_res = self.x_array[i]
            elif residual_noiseless[i] == top_6[3]:
                max_PE_4th_res = self.x_array[i]
            elif residual_noiseless[i] == top_6[4]:
                max_PE_5th_res = self.x_array[i]
            elif residual_noiseless[i] == top_6[5]:
                max_PE_6th_res = self.x_array[i]
            elif residual_noiseless[i] == bottom_6[0]:
                min_PE_1st_res = self.x_array[i]
            elif residual_noiseless[i] == bottom_6[1]:
                min_PE_2nd_res = self.x_array[i]
            elif residual_noiseless[i] == bottom_6[2]:
                min_PE_3rd_res = self.x_array[i]
            elif residual_noiseless[i] == bottom_6[3]:
                min_PE_4th_res = self.x_array[i]
            elif residual_noiseless[i] == bottom_6[4]:
                min_PE_5th_res = self.x_array[i]
            elif residual_noiseless[i] == bottom_6[5]:
                min_PE_6th_res = self.x_array[i]
        top_6_PE.append(max_PE_1st_res)
        top_6_PE.append(max_PE_2nd_res)
        top_6_PE.append(max_PE_3rd_res)
        top_6_PE.append(max_PE_4th_res)
        top_6_PE.append(max_PE_5th_res)
        top_6_PE.append(max_PE_6th_res)
        bottom_6_PE.append(min_PE_1st_res)
        bottom_6_PE.append(min_PE_2nd_res)
        bottom_6_PE.append(min_PE_3rd_res)
        bottom_6_PE.append(min_PE_4th_res)
        bottom_6_PE.append(min_PE_5th_res)
        bottom_6_PE.append(min_PE_6th_res)

        pass_cond_max = 0
        pass_cond_min = 0
        new_max_PE = 0
        new_min_PE = 0
        new_max_res = 0
        new_min_res = 0
        upper = PE_guess[0] + 2
        lower = PE_guess[0] - 2
       
        for i in range(len(top_6_PE)):
            
            if top_6_PE[i] > upper or top_6_PE[i] < lower:
                if pass_cond_max > 0:
                    pass
                else:
                    new_max_PE = top_6_PE[i]
                    new_max_res = top_6[i]
                    pass_cond_max +=1
            if bottom_6_PE[i] > upper or bottom_6_PE[i] < lower:
                if pass_cond_min > 0:
                    pass
                else:
                    new_min_PE = bottom_6_PE[i]
                    new_min_res = bottom_6[i]
                    pass_cond_min += 1


        
        

        max_peak = amp_guess[0]
        
        peak_location = 0
        for i in amp_guess: #Finding the location of the maximum amplitude peak
            if i >= max_peak:
                max_peak_location = peak_location
            peak_location += 1
        print("MINIMUM PEAK RESDIUAL", new_min_res, "-30%",-0.3*amp_guess[max_peak_location])
        print("MAXIMUM PEAK RESDIUAL", new_max_res, "15%",0.15*amp_guess[max_peak_location])
        #Check to remove peak. Won't remove if only one peak    
        if len(PE_guess) > 1:    
            if new_min_res < -0.3*amp_guess[max_peak_location] and new_min_res != 0: #NEED TO FIND A BETTER CONDITION FOR ADDING AND REMOVING PEAKS 
                #Do we want to check where the peak is located and make some evaluation from this using new_min_PE?
                print("Data is overfit. Removing peak to try to improve fitting")
                self.removePeak()

        if new_max_res > 0.15*amp_guess[max_peak_location]: #NEED TO FIND A BETTER CONDITION FOR ADDING AND REMOVING PEAKS 

            new_peak_PE = new_max_PE
            new_peak_amp = new_max_res
            print("NEW PEAK PE", new_peak_PE, "NEW PEAK AMPLITUDE", new_peak_amp)
            self.addPeak(new_peak_PE, new_peak_amp)
        else:
            print("NO NEW PEAK ADDED")
            

        
    

    

     

       


    #THIS FUNCTION IS ONLY WORKING FOR VOIGT RIGHT NOW. IMPROPER ADDITION OF THE NEW PARAMETERS WITH OTHER TYPES
    #Where do we call this function? --> Right now it is only functionable in generateFirstGen()
    def addPeak(self, new_PE, new_amp): #New function to be called after so many generations (20?). Used before next generation is created. 
        
        
        PE_guess.append(new_PE) #Want to use reisdual to determine where to add new peak?
        #Condition if the new peak trying to be added is too close to another peak --> Instead adjust amp of that peak
        #How to use reisudal to find these parameters? SOS? Br? etc. --> Take on same values as other peak but give a wider range?
      
        amp_guess.append(new_amp)
        sigma_guess.append(0.5)
        gamma_guess.append(0.25)

        amp_limited.append(False)
        PE_limited.append(False)
        gamma_limited.append(False)
        sigma_limited.append(False)
        amp_min_max = new_amp*0.2
        amp_range_min.append(-amp_min_max)
        amp_range_max.append(amp_min_max)
        amp_range_delta.append(0.05)

        PE_range_min.append(-0.2)
        PE_range_max.append(0.2)
        PE_range_delta.append(0.01)

        #Allowing for larger sigma and gamma range --> Need to figure out how to get a better estimate on these parameters from residual
        sigma_range_min.append(-0.5)
        sigma_range_max.append(0.5)
        sigma_range_delta.append(0.001)
       
        gamma_range_min.append(-0.25)
        gamma_range_max.append(0.25)
        gamma_range_delta.append(0.001)

        
        
        peak_type.append(peak_type[0]) #Appending whatever is the 1st input peak type --> Want to change in the future to have the algorithm decide what is the best peak type
        self.peak_type = peak_type
      

        i = len(PE_guess)

        PE_correlated.append(i)
        amp_correlated.append(i)
        gamma_correlated.append(i)
        sigma_correlated.append(i)

        #How to allow for peak correlation for added peak? Correlate if other peaks have correlation?
        PE_correlated_mult.append(1)
        amp_correlated_mult.append(1)
        sigma_correlated_mult.append(1)
        gamma_correlated_mult.append(1)

       
        #Take on the same values as the first peak? Maximum amp peak?
        branching_ratio.append(branching_ratio[0])
        spinOrbitSplit_guess.append(spinOrbitSplit_guess[0])
        gamma_CK_guess.append(gamma_guess[len(PE_guess)-1]) #taking in same gamma value as the newest gamma added
        is_singlet.append(is_singlet[0])
        is_coster_kronig.append(is_coster_kronig[0])


        npaths = len(PE_guess)
       


        #HAVE TO RECALL ALL THIS SO THAT WE CAN CALL PARS RANGE AGAIN
        photoline_select = self.photoelectronLine
        transitionLine_select = self.transitionLine
        N = 10
        y_left = self.y_array[:N]
        y_right = self.y_array[-N:]
        x_left = self.x_array[:N]
        x_right = self.x_array[-N:]

        self.y_left_avg = sum(y_left)/N
        x_left_avg = sum(x_left)/N
        self.y_right_avg = sum(y_right)/N
        x_right_avg = sum(x_right)/N
        if self.y_left_avg > self.y_right_avg:
            #Should we scale baseline so that the last point (lowest BE/highest KE) is subtratced off the y values (goes to zero) and set highest possible baseline value to the left side of the data?
            baseline_range = [-(self.y_left_avg+np.round(self.y_left_avg*0.2, 2)),self.y_left_avg+np.round(self.y_left_avg*0.05, 2),1] #Changed minimum to allow for background lower. Now allows for 20% below avg. and 5 % above
           
        else:
            baseline_range = [-(self.y_left_avg+np.round(self.y_left_avg*0.2, 2)),self.y_right_avg+np.round(self.y_right_avg*0.05, 2),1]

        amp_range = []
        PE_range = []
        sigma_range = []
        gamma_range = []

        max_peak = amp_range_max[0] #Finding the peak with the largest amplitude and giving it a larger PE range
        #In the future we will need to have the algorithm check the elemental peak assingment here and give it a range based on a database of collected peak information
        '''
        peak_location = 0
        for i in amp_range_max:
            if i >= max_peak:
                max_peak_location = peak_location
            peak_location += 1
        PE_range_min[max_peak_location] = -0.5
        PE_range_max[max_peak_location] = +0.5
        '''

        for i in range(len(PE_guess)): #Creating 2D arrays of range values for each peaks PE and amplitude

            if amp_limited[i] == True:
                amp_range.append([-0.00001, 0.00001, 0.00001]) #Making the range basically zero if the parameter is 'fixed'
            else:
                amp_range.append([amp_range_min[i], amp_range_max[i], amp_range_delta[i]])


            if PE_limited[i] == True:
                
                PE_range.append([-0.00001, 0.00001, 0.00001]) #Making the range basically zero if the parameter is 'fixed'
            else:
                PE_range.append([PE_range_min[i], PE_range_max[i], PE_range_delta[i]])
        


            if sigma_limited[i] == True:
                sigma_range.append([-0.00001, 0.00001, 0.00001]) #Making the range basically zero if the parameter is 'fixed'
            else:
                sigma_range.append([sigma_range_min[i], sigma_range_max[i], sigma_range_delta[i]])



            if gamma_limited[i] == True:
                gamma_range.append([-0.00001, 0.00001, 0.00001]) #Making the range basically zero if the parameter is 'fixed'
            else:
                gamma_range.append([gamma_range_min[i], gamma_range_max[i], gamma_range_delta[i]])


        #REDFINE parse_range after appending everything needed to add a peak

        #Recalling self.pars_range after fitting has already begun seems to restart the guesses for values to the values the algorithm has found and creates new ranges around those values
        self.pars_range = {
            'Photon Energy': PE_range,
            'PE':PE_guess,
            'PE_limited' :PE_limited,
            'PE_correlated' :PE_correlated,
            'PE_correlated_mult' :PE_correlated_mult,
            'Gaussian': sigma_range,
            'Sigma': sigma_guess,
            'sigma_limited' :sigma_limited,
            'sigma_correlated' :sigma_correlated,
            'sigma_correlated_mult' :sigma_correlated_mult,
            'Lorentzian': gamma_range,
            'Gamma': gamma_guess,
            'gamma_limited' :gamma_limited,
            'gamma_correlated' :gamma_correlated,
            'gamma_correlated_mult' :gamma_correlated_mult,
            'Amplitude' : amp_range,
            'Amp': amp_guess,
            'amp_limited' :amp_limited,
            'amp_correlated' :amp_correlated,
            'amp_correlated_mult' :amp_correlated_mult,
            'Asymmetry' : asymmetry_range,
            'Asymmetry Doniach-Sunjic' : asymmetryDoniach_range,
            'k_range' : k_range,
            'Background' : background_range,
            #'Shirley Background' : background_shir_range,
            'CTou3' : CTou3_range,
            'DTou3' : DTou3_range,
            'Slope' : slope_range,
            'Exponential Amplitude' : exp_amp_range,
            'Decay' : exp_decay_range,
            'npeaks' : npaths,
            'baseline' : baseline_range,
            'branching_ratios' : branching_ratio,
            'is_singlet' : is_singlet,
            'is_coster_kronig' : is_coster_kronig,
            'Lorentzian Coster-Kronig' : gamma_CK_range,
            'Gamma Coster-Kronig' : gamma_CK_guess,
            'spinOrbitSplitting' : spinOrbitSplit_range,
            'spinOrbitSplit' : spinOrbitSplit_guess,
            'photoline' : photoline_select,
            'transitionline' :transitionLine_select
        }

    def removePeak(self): #Removing peak from array. Right now it just removes the last input --> Need to change that 
        
        
        PE_guess.pop() 
      
        amp_guess.pop()
        sigma_guess.pop()
        gamma_guess.pop()

        amp_limited.pop()
        PE_limited.pop()
        gamma_limited.pop()
        sigma_limited.pop()
        
        amp_range_min.pop()
        amp_range_max.pop()
        amp_range_delta.pop()

        PE_range_min.pop()
        PE_range_max.pop()
        PE_range_delta.pop()

       
        sigma_range_min.pop()
        sigma_range_max.pop()
        sigma_range_delta.pop()

        gamma_range_min.pop()
        gamma_range_max.pop()
        gamma_range_delta.pop()

        peak_type.pop()
        self.peak_type = peak_type
      

       

        PE_correlated.pop()
        amp_correlated.pop()
        gamma_correlated.pop()
        sigma_correlated.pop()
       
        PE_correlated_mult.pop()
        amp_correlated_mult.pop()
        sigma_correlated_mult.pop()
        gamma_correlated_mult.pop()


        branching_ratio.pop()
        spinOrbitSplit_guess.pop()
        gamma_CK_guess.pop()
        is_singlet.pop()
        is_coster_kronig.pop()
        npaths = len(PE_guess)
       


        #HAVE TO RECALL ALL THIS SO THAT WE CAN CALL PARS RANGE AGAIN
        photoline_select = self.photoelectronLine
        transitionLine_select = self.transitionLine
        N = 10
        y_left = self.y_array[:N]
        y_right = self.y_array[-N:]
        x_left = self.x_array[:N]
        x_right = self.x_array[-N:]

        self.y_left_avg = sum(y_left)/N
        x_left_avg = sum(x_left)/N
        self.y_right_avg = sum(y_right)/N
        x_right_avg = sum(x_right)/N
        if self.y_left_avg > self.y_right_avg:
            #Should we scale baseline so that the last point (lowest BE/highest KE) is subtratced off the y values (goes to zero) and set highest possible baseline value to the left side of the data?
            baseline_range = [-(self.y_left_avg+np.round(self.y_left_avg*0.2, 2)),self.y_left_avg+np.round(self.y_left_avg*0.05, 2),1] #Changed minimum to allow for background lower. Now allows for 20% below avg. and 5 % above
           
        else:
            baseline_range = [-(self.y_left_avg+np.round(self.y_left_avg*0.2, 2)),self.y_right_avg+np.round(self.y_right_avg*0.05, 2),1]

        amp_range = []
        PE_range = []
        sigma_range = []
        gamma_range = []

        max_peak = amp_range_max[0] #Finding the peak with the largest amplitude and giving it a larger PE range
        #In the future we will need to have the algorithm check the elemental peak assingment here and give it a range based on a database of collected peak information
        peak_location = 0
        for i in amp_range_max:
            if i >= max_peak:
                max_peak_location = peak_location
            peak_location += 1
        PE_range_min[max_peak_location] = -0.5
        PE_range_max[max_peak_location] = +0.5

        for i in range(len(PE_guess)): #Creating 2D arrays of range values for each peaks PE and amplitude
         
            if amp_limited[i] == True:
                amp_range.append([-0.00001, 0.00001, 0.00001]) #Making the range basically zero if the parameter is 'fixed'
            else:
                amp_range.append([amp_range_min[i], amp_range_max[i], amp_range_delta[i]])



            if PE_limited[i] == True:
                
                PE_range.append([-0.00001, 0.00001, 0.00001]) #Making the range basically zero if the parameter is 'fixed'
            else:
                PE_range.append([PE_range_min[i], PE_range_max[i], PE_range_delta[i]])
        


            if sigma_limited[i] == True:
                sigma_range.append([-0.00001, 0.00001, 0.00001]) #Making the range basically zero if the parameter is 'fixed'
            else:
                sigma_range.append([sigma_range_min[i], sigma_range_max[i], sigma_range_delta[i]])



            if gamma_limited[i] == True:
                gamma_range.append([-0.00001, 0.00001, 0.00001]) #Making the range basically zero if the parameter is 'fixed'
            else:
                gamma_range.append([gamma_range_min[i], gamma_range_max[i], gamma_range_delta[i]])


        #REDFINE parse_range after appending everything needed to add a peak
        self.pars_range = {
            'Photon Energy': PE_range,
            'PE':PE_guess,
            'PE_limited' :PE_limited,
            'PE_correlated' :PE_correlated,
            'PE_correlated_mult' :PE_correlated_mult,
            'Gaussian': sigma_range,
            'Sigma': sigma_guess,
            'sigma_limited' :sigma_limited,
            'sigma_correlated' :sigma_correlated,
            'sigma_correlated_mult' :sigma_correlated_mult,
            'Lorentzian': gamma_range,
            'Gamma': gamma_guess,
            'gamma_limited' :gamma_limited,
            'gamma_correlated' :gamma_correlated,
            'gamma_correlated_mult' :gamma_correlated_mult,
            'Amplitude' : amp_range,
            'Amp': amp_guess,
            'amp_limited' :amp_limited,
            'amp_correlated' :amp_correlated,
            'amp_correlated_mult' :amp_correlated_mult,
            'Asymmetry' : asymmetry_range,
            'Asymmetry Doniach-Sunjic' : asymmetryDoniach_range,
            'k_range' : k_range,
            'Background' : background_range,
            #'Shirley Background' : background_shir_range,
            'CTou3' : CTou3_range,
            'DTou3' : DTou3_range,
            'Slope' : slope_range,
            'Exponential Amplitude' : exp_amp_range,
            'Decay' : exp_decay_range,
            'npeaks' : npaths,
            'baseline' : baseline_range,
            'branching_ratios' : branching_ratio,
            'is_singlet' : is_singlet,
            'is_coster_kronig' : is_coster_kronig,
            'Lorentzian Coster-Kronig' : gamma_CK_range,
            'Gamma Coster-Kronig' : gamma_CK_guess,
            'spinOrbitSplitting' : spinOrbitSplit_range,
            'spinOrbitSplit' : spinOrbitSplit_guess,
            'photoline' : photoline_select,
            'transitionline' :transitionLine_select
        }
       
        
        




    def generateIndividual(self):
        """
        Generate singular individual
        """
        
        ind = Individual(self.backgrounds,peak_type,self.scale_var,self.pars_range)
        
        return ind

    def generateFirstGen(self):
        print("generating first gen")
        self.Populations=[]

        #ind = Individual(self.backgrounds,peak_type,self.scale_var,self.pars_range)
        #self.addPeak(ind)

        
        for i in range(self.npops):
           
            self.Populations.append(self.generateIndividual())
      
        self.eval_Population()
        self.globBestFit = self.sorted_population[0]

        print("First gen generated")
    # @profile
    def fitness(self,indObj):
        """
        Evaluate fitness of a individual
        """
        loss = 0
        Individual = indObj
        yTotal = np.zeros(len(self.x_slice))
        self.residual = np.zeros(len(self.x_slice))
        
        #HOW TO MAKE SCALE VALUE SEEN HERE?
        yTotal = Individual.getFit(self.x_array,self.y_array, self.backgrounds)
       

        #Segmenting data to see where y falls off 
        
        big_diff = 0

        for i in range(len(self.x_array)-5):
            y_b4 = self.y_array[i]
           
            y_after = self.y_array[i+5]
            
            y_diff = y_b4 - y_after
            if self.x_array[0] < self.x_array[-1]: #KE condition

                if y_diff > 0:
                    new_diff = y_diff
                    if new_diff > big_diff:
                        big_diff = new_diff
                        location = i
            else:

                if y_diff < 0:
                    new_diff = y_diff
                    if new_diff < big_diff:
                        big_diff = new_diff
                        location = i
           

                #print("X B4", self.x[location], "X AFTER", self.x[location+40])
   
        y_low_PE_penalty = 0
        for k in range(location, len(self.x_array)-1):
            if (yTotal[k]- self.y_array[k]) > 0:
                y_low_PE_penalty += (yTotal[k]- self.y_array[k])*5000 #Dont think we need this penalty since we are accomadating for it in the bkgn_penalty

        for j in range(len(self.x_array)):

           # loss = loss + (yTotal[j]*self.x_array[j]**2 - self.y_array[j]* self.x_array[j]**2 )**2
            #loss = loss + (((yTotal[j]- self.y_array[j])**2)*self.y_array[j]
            #loss = loss + (((yTotal[j]- self.y_array[j])**2))*np.sqrt(self.y_array[j])

             #include a penalty for going outside of the spectral envelope
            penalty = 0
            bkgn_penalty = 0
            overfit_penalty = 0
            difference = (yTotal[j] - self.y_array[j])
            if difference > 0:
                overfit_penalty = difference*10000 #Adding penalty for going above the data
            sigma = np.sqrt(abs(self.y_array[j]))
          
            N = 30

            if sum(yTotal[:N])/N > self.y_left_avg: #Penalty for background going above the data on left side of plot
                diff = sum(yTotal[:N])/N - self.y_left_avg
                bkgn_penalty += 300000*sigma*diff

            if sum(yTotal[:N])/N < self.y_left_avg: #Penalty for background going below the data on left side of plot
                diff = abs(sum(yTotal[:N])/N - self.y_left_avg)
                bkgn_penalty += 300000*sigma*diff

            if sum(yTotal[-N:])/N > self.y_right_avg: #Penalty for background going above the data on right side of plot
                diff = sum(yTotal[-N:])/N - self.y_right_avg
                bkgn_penalty += 30000*sigma*diff

            if sum(yTotal[-N:])/N < self.y_right_avg: #Penalty for going below the data on right side of plot
                diff = abs(sum(yTotal[-N:])/N - self.y_right_avg)
                bkgn_penalty += 300000*sigma*diff

            if difference > 0:
                if difference >= 0.5*sigma:
                    penalty = 1000*sigma #Dont know if we need this is we already are taking into account the overfit_penalty
           
            loss = loss + (difference**2)*sigma + penalty*sigma + bkgn_penalty #+ overfit_penalty + y_low_PE_penalty
          
            self.residual[j] = ((yTotal[j]- self.y_array[j])) 
       
      
         # if loss == np.nan:
            # print(individual[0].verbose())
    
        return loss, yTotal
   
    def eval_Population(self):
        """
        Evalulate populations
        """
        
        score = []
        populationPerf = {}
        best_y = []
        temp_temp_score = 10000000000000000000000000
        yPerf = {}
        self.nan_counter = 0
        
        for i,individual in enumerate(self.Populations):
            temp_score, bestY = self.fitness(individual)
            
            #NEW PLAN FOR UPGRADING XES NEO:
            #Take residual from fitness and use it to determine change in range parameters or addition/sutraction of a peak
            #This should come into play only after so many generations without change or with very large residual
            #May need to add in another function that would add/remove a peak --> Get BE from residual idk how to get sigma/gamma/amp?
            #Another function for adjusting parameters --> Algorithm chooses between these two options based on some criteria
            
            #Finding the best y values from the best individual in the fit so far
            if temp_score < temp_temp_score:
                best_y = bestY
                temp_temp_score = temp_score
            # Calculate the score, if encounter nan, discard and generate new individual later
            if np.isnan(temp_score): #Change to be if residual is bad AND numGenSinceImproved > 10% of generation size then .... add/remove peak OR adjust parameters
                self.nan_counter +=1
            else:
                score.append(temp_score)
                populationPerf[individual] = temp_score
                
                
               
        self.sorted_population = sorted(populationPerf.items(), key=operator.itemgetter(1), reverse=False) #Get an error here if y values go below zero for some reason
        
        '''
        for a,b in self.sorted_population:
            print(str(b) + " " + str(a.get_params()))
        '''
        ''' Debugging again
        for i in range(len(self.sorted_population)):
            print(self.sorted_population[i][0].get_peak(0))
        '''
        self.currBestFit = self.sorted_population[0]
        self.currBestY = best_y
      

        return score

    
    def next_generation(self):
        """
        Calculate next generations

        """
        self.st = time.time()
        # ray.init()
        self.logger.info("---------------------------------------------------------")
        self.logger.info(datetime.datetime.fromtimestamp(self.st).strftime('%Y-%m-%d %H:%M:%S'))
        self.logger.info(f"{bcolors.BOLD}Gen: {bcolors.ENDC}{self.genNum+1}")

        self.genNum += 1

        # Evaluate Fittness
        score = self.eval_Population()
        self.bestDiff = abs(self.globBestFit[1]-self.currBestFit[1])
        if self.currBestFit[1] < self.globBestFit[1]:
            self.globBestFit = self.currBestFit
            self.numGenSinceImproved = 0
        else:
            self.numGenSinceImproved += 1

        
        with np.printoptions(precision=5, suppress=True):
            self.logger.info("Different from last best fit: " +str(self.bestDiff))
            self.logger.info("Number of Generations since improved: " + str(self.numGenSinceImproved))
            self.logger.info(bcolors.BOLD + "Best fit: " + bcolors.OKBLUE + str(self.currBestFit[1]) + bcolors.ENDC)
            self.logger.info("Best fit combination:\n" + str((self.sorted_population[0][0].get_params())))
            self.logger.info(bcolors.BOLD + "History Best: " + bcolors.OKBLUE + str(self.globBestFit[1]) +bcolors.ENDC)
            #self.logger.info("NanCounter: " + str(self.nan_counter))
            self.logger.info("History Best Indi:\n" + str((self.globBestFit[0].get_params())))

        nextBreeders = self.selectFromPopulation()
        self.logger.info("Number of Breeders: " + str(len(self.parents)))
        self.logger.info("DiffCounter: " + str(self.diffCounter))
        self.logger.info("Diff %: " + str(self.diffCounter / self.genNum))
        self.logger.info("Mutation Chance: " + str(self.mut_chance))
        self.mutatePopulation()
        self.createChildren()

        
        self.et = timecall()
        self.tdiff = self.et - self.st
        self.tt = self.tt + self.tdiff
        self.logger.info("Time: "+ str(round(self.tdiff,5))+ "s")
       
    def mutatePopulation(self):
        """
        # Mutation operators
        # 0 = original: generated a new versions:
        # 1 = mutated every genes in the total populations
        # 2 = mutated genes inside population based on secondary probability

        # TODO:
            options 2 and 3 needs to reimplmented
        """
        self.nmutate = 0

        if self.mut_opt  == 0:
            # Rechenberg mutation
            if self.genNum > 20:
                if self.bestDiff < 0.1:
                    self.diffCounter += 1
                else:
                    self.diffCounter -= 1
                if (abs(self.diffCounter)/ float(self.genNum)) > 0.2:
                    self.mut_chance += 0.5
                    self.mut_chance = abs(self.mut_chance)
                elif (abs(self.diffCounter) / float(self.genNum)) < 0.2:
                    self.mut_chance -= 0.5
                    self.mut_chance = abs(self.mut_chance)


        for i in range(self.npops):
            if random.random()*100 < self.mut_chance:
                self.nmutate += 1
                self.Populations[i] = self.mutateIndi(i)

        self.logger.info("Mutate Times: " + str(self.nmutate))


    def mutateIndi(self,indi):
        """
        Generate new individual during mutation operator
        """
        if self.mut_opt == 0:
            # Create a new individual with Rechenberg
            newIndi = self.generateIndividual()
        # Random pertubutions
        if self.mut_opt == 1:
            # Random Pertubutions
            self.Populations[indi].mutate_(self.mut_chance)
            newIndi = self.Populations[indi]
            # Mutate every gene in the Individuals

        if self.mut_opt == 2:
            # initalize_variable:
            self.nmutate_success = 0
            og_indi = copy.deepcopy(self.Populations[indi])
            og_score = self.fitness(og_indi)
            mut_indi = copy.deepcopy(self.Populations[indi])
            mut_indi.mutate_(self.mut_chance)
            mut_score = self.fitness(mut_indi)

            with np.errstate(divide='raise', invalid='raise'):
                try:
                    t_bot = (np.log(1-(self.genNum/self.ngen)+self.tol))
                except FloatingPointError:
                    print(self.genNum)
                    print(self.ngen)
                    print(1-(self.genNum/self.ngen))
                    t_bot = (np.log(1-(self.genNum/self.ngen)+self.tol))

            T = - self.bestDiff/t_bot
            if mut_score < og_score:
                self.nmutate_success = self.nmutate_success + 1;
                newIndi = mut_indi
            elif np.exp(-(mut_score-og_score)/(T+self.tol)) > np.random.uniform():

                self.nmutate_success = self.nmutate_success + 1;
                newIndi = mut_indi
            else:
                newIndi = og_indi

        if self.mut_opt == 3: #Not working --> Replace with DE?
            def delta_fun(t,delta_val):
                rnd = np.random.random()
                return delta_val*(1-rnd**(1-(t/self.ngen))**5)

            og_indi = copy.deepcopy(self.Populations[indi])
            og_data = og_indi.get_var() #error --> no value get_var
            for i,path in enumerate(og_data):
                print(i,path)
                arr = np.random.randint(2,size=3)
                for j in range(len(arr)):
                    new_path = []
                    val = path[j]
                    if arr[j] == 0:
                        UP = self.pathrange_Dict[i].get_lim()[j+1][1]
                        del_val = delta_fun(self.genNum,UP-val)
                        val = val + del_val
                    if arr[j] == 1:
                        LB = self.pathrange_Dict[i].get_lim()[j+1][0]
                        del_val = delta_fun(self.genNum,val-LB)
                    new_path.append(val)
                self.Populations[indi].set_path(i,new_path[0],new_path[1],new_path[2])
        if self.mut_opt == 4:
            newIndi = self.generateIndividual(self.bestE0)
        return newIndi

    def selectFromPopulation(self):
        self.parents = []

        select_val = np.minimum(self.n_bestsam,len(self.sorted_population))
        self.n_recover = 0
        if len(self.sorted_population) < self.n_bestsam:
            self.n_recover = self.n_bestsam - len(self.sorted_population)
        for i in range(select_val):
            self.parents.append(self.sorted_population[i][0])

    def crossover(self,individual1, individual2):
        """
        Uniform Cross-Over, 50% percentage chance
        """
     
        child = self.generateIndividual()

        individual1_path = individual1.get_params()
        individual2_path = individual2.get_params()

        #print("Ind 1 : " + str(individual1_path))
        #print("Ind 2 : " + str(individual2_path))
        temp_path = []
        dividers = [] # markers where the strings are in the list of params, this indicates where the array switches to a new peak or background
        #crossover for peak vars
        for j in range(len(individual1_path)):
            if (isinstance(individual1_path[j],str)):
                dividers.append(j)
            if np.random.randint(0,2) == True:
                temp_path.append(individual1_path[j])
            else:
                temp_path.append(individual2_path[j])
            '''
        for j in range(1):
            if np.random.randint(0,2) == True:
                temp_path.append(individual1_path[1][j])
            else:
                temp_path.append(individual2_path[1][j])
        '''
        #print("Temp Path: " + str(temp_path))
        temp_peak = []
        #print(temp_path)
        divider = 0
        peakNum = 0
        bkgnNum = 0
        for k in range(len(dividers)):
            for j in range(divider,dividers[k]+1):
                temp_peak.append(temp_path[j])
            if i < self.npaths:
                
                #print()
                #print("Child pre-write: " + str(child.get_params()))
                #print("temp peak : " + str(temp_peak))
                if child.setPeak(peakNum,temp_peak) == -1:
                    if bkgnNum<len(background_type):
                        #print("Bkgn")
                        child.setBkgn(bkgnNum,temp_peak)
                        bkgnNum += 1
                else:
                    #print("wrote peak")
                    peakNum +=1
                #print("Child after write")
                #print(child.get_params())
                #print()
                temp_peak = []
            divider = j + 1

        #print("Child : " + str(child.get_params()))
        '''
        child.setPeak(i,temp_path[0],temp_path[1],temp_path[2],temp_path[3])
        child.get_background(0).set_k(temp_path[4])
        '''
        '''
        print(temp_path)
        print("Child:")
        print(child.get_params())
        exit()
        '''
        return child

    def createChildren(self):
        """
        Generate Children
        """
        self.nextPopulation = []
        # --- append the breeder ---
        for i in range(len(self.parents)):
            self.nextPopulation.append(self.parents[i])
        # print(len(self.nextPopulation))
        # --- use the breeder to crossover
        # print(abs(self.npops-self.n_bestsam)-self.n_lucksam)

        for i in range(abs(self.npops-self.n_bestsam)-self.n_lucksam):
            par_ind = np.random.choice(len(self.parents),size=2,replace=False)
            child = self.crossover(self.parents[par_ind[0]],self.parents[par_ind[1]])
            self.nextPopulation.append(child)
        # print(len(self.nextPopulation))

        for i in range(self.n_lucksam):
            self.nextPopulation.append(self.generateIndividual())
        # print(len(self.nextPopulation))

        for i in range(self.n_recover):
            self.nextPopulation.append(self.generateIndividual())

        # for i in range(self.nan_counter):
        #     self.nextPopulation.append(self.generateIndividual())

        random.shuffle(self.nextPopulation)
        self.Populations = self.nextPopulation

    def run_verbose_start(self):
        self.logger.info("-----------Inputs File Stats---------------")
        self.logger.info(f"{bcolors.BOLD}File{bcolors.ENDC}: {self.data_file}")
        #self.logger.info(f"{bcolors.BOLD}File Type{bcolors.ENDC}: {self.data_obj._ftype}")
        self.logger.info(f"{bcolors.BOLD}File{bcolors.ENDC}: {self.output_path}")
        self.logger.info(f"{bcolors.BOLD}Population{bcolors.ENDC}: {self.npops}")
        self.logger.info(f"{bcolors.BOLD}Num Gen{bcolors.ENDC}: {self.ngen}")
        self.logger.info(f"{bcolors.BOLD}Mutation Opt{bcolors.ENDC}: {self.mut_opt}")
        self.logger.info("-------------------------------------------")

    def run_verbose_end(self):
        self.logger.info("-----------Output Stats---------------")
        # self.logger.info(f"{bcolors.BOLD}Total)
        self.logger.info(f"{bcolors.BOLD}Total Time(s){bcolors.ENDC}: {round(self.tt,4)}")
        self.logger.info("-------------------------------------------")

    def run(self, data_peak_add):
        self.run_verbose_start()
        self.historic = []
        self.historic.append(self.Populations)
        count = 0
        peak_add = 0
        before_best_fit = 0
        num_peaks = len(PE_guess)
        for i in range(self.ngen):
           
            temp_gen = self.next_generation()
        
       
            #Peak addition/removal determined by user checking selection button in GUI
            
            if peak_add_remove == True:
                
            
                if count == 5*len(PE_guess): #Condition for adding new peak. Does so after so many generations --> Maybe make peak num dependent 10*numPeaks --> Need more time to analyze more peaks
                
                    if peak_add > 0: 
                    
                        if before_best_fit < self.currBestFit[1]: #Check to make sure that the added peak improved the fit
                            #REMOVE PEAK THAT WAS ADDED
                            self.removePeak()
                            print("No improvement from peak addition. Added peak removed")
                    
                
                    self.scanResidual()
                    if len(PE_guess) > num_peaks:
                        num_peaks = len(PE_guess)
                        self.generateFirstGen()
                        if os.path.exists(self.file_data): os.remove(self.file_data)
                    elif len(PE_guess) < num_peaks:
                        num_peaks = len(PE_guess)
                        self.generateFirstGen()
                        if os.path.exists(self.file_data): os.remove(self.file_data) #CAN ADD NEW PEAKS HERE BUT IT WILL NEED TO DESTROY THE OLD OUTPUT FILE AND CREATE A NEW ONE SO THAT THE NUMBER OF COLUMNS ARE ALL THE SAME FOR ANALYSIS
                    before_best_fit = self.currBestFit[1]
                    count = 0
                    peak_add += 1
            
            
            self.output_generations()
            count += 1
           

        #print(self.globBestFit[0].getFit(self.x_array,self.y_array))
        self.run_verbose_end()
        # test_y = self.export_paths(self.globBestFit[0])
        # plt.plot(self.data_obj.get_raw_data()[:,0],self.data_obj.get_raw_data()[:,1],'b-.')
        # plt.plot(self.x_slice,self.y_slice,'o--',label='data')
        # plt.plot(self.x_slice,test_y,'r--',label='model')
        # plt.legend()
        # plt.show()

    def export_paths(self,indObj):
        area_list=[]
        Individual = indObj.get()

        yTotal = np.zeros(len(self.x_slice))
        plt.figure()
        for i,paths in enumerate(Individual):
            y = paths.getY()

            yTotal += y
            # area = np.trapz(y.flatten(),x=self.x_slice.flatten())
            # component = paths.get_func(self.x_slice).reshape(-1,1)

            # area_list.append(area)

        Total_area = np.sum(area_list)
        return yTotal

    def output_generations(self):
        """
        Output generations result into two files
        """
        try:
            f1 = open(self.file,"a")
            f1.write(str(self.genNum) + "," + str(self.tdiff) + "," +
                str(self.currBestFit[1]) + "," + str(self.currBestFit[0].get_params()) +")," +
                str(self.globBestFit[1]) + "," + str(self.globBestFit[0].get_params()) +"\n")
        finally:
            f1.close()
        try:
            f2 = open(self.file_data,"a")
            write = csv.writer(f2)
            bestFit = self.globBestFit[0]
            #write.writerow((bestFit[i][0], bestFit[i][1], bestFit[i][2]))
            str_pars = bestFit.get_params(for_output_file = True)
            write.writerow(str_pars)
            f2.write("#################################\n")
        finally:
            f2.close()

    def __init__(self,scale_var = False,data_KE = False, data_XES = False, data_peak_add = False):
        """
        Steps to Initalize XES
            XES
        """
        
        self.scale_var = scale_var
        # initialize params
        self.initialize_params()
        # variables
        self.initialize_variable()
        # initialze file paths
        self.initialize_file_path()
        # initialize range
        self.initialize_range(scale_var,data_KE, data_XES)
        # Generate first generation
        self.generateFirstGen()
      

        self.run(data_peak_add)

def main():
    XES_GA()

if __name__ == "__main__":
    main()



from xes_neo.xes_fit import background
from xes_neo.input_arg import *
from xes_neo.helper import str_to_bool
# ------------------------------------
# Andy Lau
# 4/22/2019
# goal : Parsed the ini into each specific documents.
#-------------------------------------

#UPDATE: ANY CHANGES IN THIS FILE WILL ONLY UPDATE WITH pip install . to the cd XES_Neo location

def split_string(var_dict,label):
	"""
	Read the path list
	"""
	# print(num_compounds)
	arr_str = var_dict[label]
	starter = []
	end = []
	k = 0
	split_str = []
	for i in arr_str:
		if i == '[':
			starter.append(k)
		elif i == ']':
			end.append(k)
		k = k + 1


	assert(len(starter) == len(end)),'Bracket setup not right.'
	# if num_compounds > 1:
	# 	assert(num_compounds == len(starter)),'Number of compounds not matched.'
	# 	assert(num_compounds == len(end)),'Number of compounds not matched.'

	# check if both are zeros, therefore the array is one 1 dimensions
	# arr_str = optional_var(var_dict,label,[],list)
	if len(starter) == 0 and len(end) == 0:
		split_str = list(arr_str.split(","))
	f_arr = []
	for i in range(len(split_str)):
		f_arr.append(float(split_str[i]))

	return f_arr

def optional_var(dict,name_var,alt_var=None,type_var=int):
	"""
	Detections of optional variables exists within input files, and
		put in corresponding default inputs parameters.
	"""
	# boolean needs special attentions
	if type_var == bool:
		if name_var in dict:
			return_var = str_to_bool(dict[name_var])
		else:
			return_var = alt_var
	elif type_var == None:
		if name_var in dict:
			return_var = dict[name_var]
		else:
			return_var = None
	else:
		if name_var in dict:
			return_var = type_var(dict[name_var])
		else:
			return_var = type_var(alt_var)
	return return_var

def optional_range(var_dict,label):
	if label not in var_dict:
		# Label is not therefore
		return_var = []
	else:
		return_var = split_string(var_dict,label)

	return return_var
def optional_range_string(var_dict, label):
    if label not in var_dict:
        return_var = []
    else:
       return_var = var_dict[label].split(',')


    return return_var


# -----
Inputs_dict = file_dict['Inputs']
Populations_dict = file_dict['Populations']
Mutations_dict = file_dict['Mutations']
Paths_dict = file_dict['Paths']
Outputs_dict = file_dict['Outputs']
# -----

# Input
data_file = Inputs_dict['data_file']
output_file = Inputs_dict['output_file']
skipLn = int(Inputs_dict['skipln'])

#data_cutoff = [float(x) for x in Inputs_dict['data_cutoff'].split(',')] -Unneccesary, remove later when we have stable build
#pathrange_file = optional_var(Inputs_dict,'pathrange_file',None,None)		-evan


# population
size_population = int(Populations_dict['population'])
number_of_generation = int(Populations_dict['num_gen'])
best_sample = int(Populations_dict['best_sample'])
lucky_few = int(Populations_dict['lucky_few'])

# Mutations
chance_of_mutation = int(Mutations_dict['chance_of_mutation'])
original_chance_of_mutation = int(Mutations_dict['original_chance_of_mutation'])
mutated_options = int(Mutations_dict['mutated_options'])

# Paths  -Should change the name of this to be peak or params or something later for readability - evan
npaths = int(Paths_dict['npeaks'])


#Rearranged to put all input guesses in order they appear in Fitting Parameters tab
PE_guess = optional_range(Paths_dict,'pe')
sigma_guess = optional_range(Paths_dict,'sigma_guess')
gamma_guess = optional_range(Paths_dict,'gamma_guess')
amp_guess = optional_range(Paths_dict,'amp_guess')



is_singlet = optional_range_string(Paths_dict,'is_singlet')
for i in range(len(is_singlet)):
	if is_singlet[i].strip() == 'True':
		is_singlet[i] = True
	else:
		is_singlet[i] = False
branching_ratio = optional_range(Paths_dict,'branching_ratio')
		
spinOrbitSplit_guess = optional_range(Paths_dict,'spinorbitsplit') #READS IT ALL IN LOWERCASE --> OTHERWISE YOU GET AN ERROR

is_coster_kronig = optional_range_string(Paths_dict,'is_coster_kronig')
for i in range(len(is_coster_kronig)):
	if is_coster_kronig[i].strip() == 'True':
		is_coster_kronig[i] = True
	else:
		is_coster_kronig[i] = False

peak_add_remove = optional_range_string(Paths_dict,'peak_adding')
scale_data = optional_range_string(Paths_dict,'scale_bool')

for i in range(len(peak_add_remove)):
	if peak_add_remove[i] == 'True':
		peak_add_remove = True
	else:
		peak_add_remove[i] = False

for i in range(len(scale_data)):
	if scale_data[i] == 'True':
		scale_data = True
	else:
		scale_data[i] = False






element = str(Paths_dict['element_select'])
photoelectronLine = str(Paths_dict['photoline_select'])
transitionLine = str(Paths_dict['transitionline_select'])

spinOrbitSplit_guess = optional_range(Paths_dict,'spinorbitsplit')
gamma_CK_range = optional_range(Paths_dict,'gamma_ck_range')
gamma_CK_guess = optional_range(Paths_dict,'gamma_ck_guess') #Same as other gamma to start


spinOrbitSplit_range = optional_range(Paths_dict,'spinorbitsplit_range')
background_type = optional_range_string(Paths_dict,'background_type')
peak_type = optional_range_string(Paths_dict,'peak_type')
#PE_range = optional_range(Paths_dict,'pe_range')
PE_range_min = optional_range(Paths_dict,'pe_range_min')
PE_range_max = optional_range(Paths_dict,'pe_range_max')
PE_range_delta = optional_range(Paths_dict,'pe_range_delta')
PE_limited = optional_range_string(Paths_dict,'pe_limited')
for i in range(len(PE_limited)):
	if PE_limited[i].strip() == 'True':
		PE_limited[i] = True
	else:
		PE_limited[i] = False
PE_correlated = optional_range_string(Paths_dict,'pe_correlated') 
for i in range(len(PE_correlated)):
	if PE_correlated[i].strip() == 'Peak #':
		PE_correlated[i] = i + 1
	else:
		PE_correlated[i] = PE_correlated[i]
PE_correlated_mult = optional_range(Paths_dict,'pe_correlated_mult')

sigma_range_min = optional_range(Paths_dict,'sigma_range_min')
sigma_range_max = optional_range(Paths_dict,'sigma_range_max')
sigma_range_delta = optional_range(Paths_dict,'sigma_range_delta')
sigma_limited = optional_range_string(Paths_dict,'sigma_limited')
for i in range(len(sigma_limited)):
	if sigma_limited[i].strip() == 'True':
		sigma_limited[i] = True
	else:
		sigma_limited[i] = False
sigma_correlated = optional_range_string(Paths_dict,'sigma_correlated') 
for i in range(len(sigma_correlated)):
	if sigma_correlated[i].strip() == 'Peak #':
		sigma_correlated[i] = i + 1
	else:
		sigma_correlated[i] = sigma_correlated[i]
sigma_correlated_mult = optional_range(Paths_dict,'sigma_correlated_mult')

gamma_range_min = optional_range(Paths_dict,'gamma_range_min')
gamma_range_max = optional_range(Paths_dict,'gamma_range_max')
gamma_range_delta = optional_range(Paths_dict,'gamma_range_delta')
gamma_limited = optional_range_string(Paths_dict,'gamma_limited')
for i in range(len(gamma_limited)):
	if gamma_limited[i].strip() == 'True':
		gamma_limited[i] = True
	else:
		gamma_limited[i] = False
gamma_correlated = optional_range_string(Paths_dict,'gamma_correlated') 
for i in range(len(gamma_correlated)):
	if gamma_correlated[i].strip() == 'Peak #':
		gamma_correlated[i] = i + 1
	else:
		gamma_correlated[i] = gamma_correlated[i]
gamma_correlated_mult = optional_range(Paths_dict,'gamma_correlated_mult')

amp_range_min = optional_range(Paths_dict,'amp_range_min')
amp_range_max = optional_range(Paths_dict,'amp_range_max')
amp_range_delta = optional_range(Paths_dict,'amp_range_delta')
amp_limited = optional_range_string(Paths_dict,'amp_limited')
for i in range(len(amp_limited)):
	if amp_limited[i].strip() == 'True':
		amp_limited[i] = True
	else:
		amp_limited[i] = False
amp_correlated = optional_range_string(Paths_dict,'amp_correlated') 
for i in range(len(amp_correlated)):
	if amp_correlated[i].strip() == 'Peak #':
		amp_correlated[i] = i + 1
	else:
		amp_correlated[i] = amp_correlated[i]
amp_correlated_mult = optional_range(Paths_dict,'amp_correlated_mult')

asymmetry_range = optional_range(Paths_dict, 'asymmetry_range')
asymmetryDoniach_range = optional_range(Paths_dict, 'asymmetrydoniach_range')
k_range = optional_range(Paths_dict,'k_range')
background_range = optional_range(Paths_dict,'background_range',)
CTou3_range = optional_range(Paths_dict,'ctou3_range')
DTou3_range = optional_range(Paths_dict,'ctou3_range')
#background_shir_range = optional_range(Paths_dict,'background_shir_range',)
slope_range = optional_range(Paths_dict,'slope_range',)
exp_amp_range = optional_range(Paths_dict,'exp_amp_range',)
exp_decay_range = optional_range(Paths_dict,'exp_decay_range',)

for i in range(len(peak_type)):
	peak_type[i].replace(" ","")

#print("Element selected is" +  ' ' + str(element))
#print("Photoelectron Line selected is" + ' ' + str(photoelectronLine))
print("Background type is " + ', '.join(background_type))
print("Peak Type is " + ', '.join(peak_type))
print("Is a Singlet: " + str(is_singlet))
print("Is a Coster-Kronig: " + str(is_coster_kronig))

print("Branching Ratio is " + str(branching_ratio))
print("Spin-Orbit Splitting Range is " + str(spinOrbitSplit_range))
print("Trying Spin-Orbit Splitting " + str(spinOrbitSplit_guess))
#print("Photon Energy Range is " + str(PE_range))
print("Photon Energy Limited " + str(PE_limited))
print("Photon Energy Correlated Peak " + str(PE_correlated))
print("Photon Energy Correlated Multipliear " + str(PE_correlated_mult))
print("Photon Energy Range Minimum is " +  str(PE_range_min))
print("Photon Energy Range Maximum is " +  str(PE_range_max))
print("Photon Energy Range Delta is " +  str(PE_range_delta))
print("Trying PE " + str(PE_guess))
print("Sigma Limited " + str(sigma_limited))
print("Sigma Correlated Peak" + str(sigma_correlated))
print("Sigma Correlated Multipliear " + str(sigma_correlated_mult))
print("Sigma Range Minimum is " + str(sigma_range_min))
print("Sigma Range Maximum is " +  str(sigma_range_max))
print("Sigma Range Delta is " +  str(sigma_range_delta))
print("Trying Sigma " + str(sigma_guess))
print("Gamma Limited " + str(gamma_limited))
print("Gamma Correlated Peak " + str(gamma_correlated))
print("Gamma Correlated Multipliear " + str(gamma_correlated_mult))
print("Gamma Range Minimum is " + str(gamma_range_min))
print("Gamma Range Maximum is " +  str(gamma_range_max))
print("Gamma Range Delta is " +  str(gamma_range_delta))
print("Trying Gamma " + str(gamma_guess))


#IDK if we should state these. They are the same as gamma
#print("Gamma Coster-Kronig Range is " + str(gamma_CK_range))
#print("Trying Gamma " + str(gamma_guess))
print("Amplitude Limited " + str(amp_limited))
print("Amplitude Correlated Peak " + str(amp_correlated))
print("Amplitude Correlated Multipliear " + str(amp_correlated_mult))
print("Amplitude Range Minimum is " +  str(amp_range_min))
print("Amplitude Range Maximum is " +  str(amp_range_max))
print("Amplitude Range Delta is " +  str(amp_range_delta))
print("Trying Amplitude " + str(amp_guess))
print("Asymmetry Range is " + str(asymmetry_range))
print("Asymmetry Doniach-Sunjic Range is " + str(asymmetryDoniach_range))
print("K range is " + str(k_range))
print("Background Range is " + str(background_range))
print("C Range is " + str(CTou3_range))
print("D Range is " + str(DTou3_range))
#print("Shirley Background Range is " + str(background_shir_range))
print("Slope Range is " + str(slope_range))
print("Exponential Amplitude Range is " + str(exp_amp_range))
print("Exponential Decay Rate Range is " + str(exp_decay_range))

print("ALLOWING PEAK ADDITION/SUBTRACTION: " + str(peak_add_remove))
print("SCALE DATA: " + str(scale_data))
# Output
printgraph = str_to_bool(Outputs_dict['print_graph'])
num_output_paths = str_to_bool(Outputs_dict['num_output_paths'])
steady_state = optional_var(Outputs_dict,'steady_state_exit',False,bool)



from xes_neo.parser2 import read_input_file
import argparse, os, sys

parser = argparse.ArgumentParser()

parser.add_argument('-i','--input',help="Submit input file")
parser.add_argument("-v","--verbose",help="output verbosity",action="store_true")
parser.add_argument("-s","--show_input",help = "show input file",action="store_true")
parser.add_argument("-t",help = "Timeing mode",action="store_true")
parser.add_argument("-d",help = "Debug mode",action="store_true")

args = parser.parse_args()

if len(sys.argv)==1:
    parser.print_help(sys.stderr)
    sys.exit(1)

if args.input!=None:
    file_path = os.path.join(os.getcwd(),args.input)
    file_dict = read_input_file(file_path)

# Read input file
if args.show_input==True and args.input != None:
    read_input_file(file_path,verbose=True)

debug_mode = args.d
timeing_mode = args.t



"""
Created 7/5/23
Created the file, left out set peak to manually set a peak and its
parameters, it seems non essential for now
-Evan Restuccia (evan@restuccias.com)
"""
from tkinter import N
from matplotlib.bezier import get_parallels
from xes_neo.xes_fit import peak,background
import numpy as np

class Individual():
    def __init__(self,backgrounds,peaks,scale_var,pars_range=''):
        """
        backgrounds (array) Array where each element is the name of the background type desired
        peaks (array) array where each element is the name of the desired peakType
        pars_range (dict) each key is the name of a parameter with a tuple that contains the range the parameter is allowed to explore
        """
       
        self.scale_var = scale_var
        self.pars_range = pars_range
        #both peaks and backgrounds are arrays of strings which represent the type of the background of the peak/bkgn
        self.nPeaks = len(peaks)

        self.nBackgrounds = len(backgrounds)
        #self.peakArr = [None]* self.nPeaks
        self.bkgnArr = [None] * self.nBackgrounds
        self.peakDict = {}
        is_singlet = pars_range['is_singlet']
        is_coster_kronig = pars_range['is_coster_kronig']

        """
        the Photon Energy needs to be personalized
        we take the range, which right now is something like (0,.2), i.e. the allowed variance in PE
        Then we add our photon energy to it to move the range to the right spot
        """
        if 'SVSC_shirley' in backgrounds:
            self.SVSC_toggle = True
            self.bkgnArr = [None] * (self.nBackgrounds-1)
        else:
            self.SVSC_toggle = False
        amp_inputs = pars_range['Amplitude'] #Declaring here so that pars_range['Amplitude'] can become a 1d array in for loop. This method isnt the best but it works
        PE_inputs = pars_range['Photon Energy']
       
        self.peakArr = [None]*len(peaks)
        sigma_inputs = pars_range['Gaussian'] 
        gamma_inputs = pars_range['Lorentzian']


        #Declare these here so that we can change pars_range to be called inside xes_fit for each peak
        full_amp_range = pars_range['Amplitude']
        full_sigma_range = pars_range['Gaussian']
        full_gamma_range = pars_range['Lorentzian']
        full_energy_range = pars_range['Photon Energy']
        
        num_peaks = pars_range['npeaks']
        
       
        for i in range(self.nPeaks): 
        
            '''
            range_key = 'Photon Energy'
            guess_key = 'PE'
            PE_range = pars_range[range_key]
            PE1,PE2 =PE_range[0],PE_range[1]
            photon_energy = pars_range[guess_key][i]
            pars_range[range_key][0],pars_range[range_key][1] = [PE_range[0] + photon_energy, PE_range[1] + photon_energy,]
            '''

            #Want to add in correlation --> make PE_value/sigma_value/gamma_value/amp_value equal to the peak correlated or to multiplication of that value by something?
            #Have range become zero for that value?
            #Not sure how this will work with the random choice value used in xes_fit --> Will the correlation only follow the entry value?

            #Redefine range values to be the correct values and make them 1D to be called in xes_fit.
            self.PE_correlated = int(pars_range['PE_correlated'][i])
           
            self.PE_correlated_mult = pars_range['PE_correlated_mult'][i]
            self.sigma_correlated = int(pars_range['sigma_correlated'][i])
            self.sigma_correlated_mult = pars_range['sigma_correlated_mult'][i]
            self.gamma_correlated = int(pars_range['gamma_correlated'][i])
            self.gamma_correlated_mult = pars_range['gamma_correlated_mult'][i]
            self.amp_correlated = int(pars_range['amp_correlated'][i])
            self.amp_correlated_mult = pars_range['amp_correlated_mult'][i]
            

            if self.PE_correlated > num_peaks:
                print("Energy Peak Correlation out of range in Parameter Ranges tab")
                exit()
            if self.sigma_correlated > num_peaks:
                print("Sigma Peak Correlation out of range in Parameter Ranges tab")
                exit()
            if self.gamma_correlated > num_peaks:
                print("Gamma Peak Correlation out of range in Parameter Ranges tab")
                exit()
            if self.amp_correlated > num_peaks:
                print("Amplitude Peak Correlation out of range in Parameter Ranges tab")
                exit()

            range_key_PE = 'Photon Energy'
            guess_key_PE = 'PE'
            PE_inputs = np.array(PE_inputs)
           
            if PE_inputs.ndim > 1: #If it is a 2D array or not
                PE_range = PE_inputs[self.PE_correlated -1] #2D array  #Takes in same range as correlated peak # -1 because i starts at zero
            else:
                PE_range = PE_inputs #1D array


            PE_value = pars_range[guess_key_PE][self.PE_correlated-1] #Takes in same input value as correlated peak
            PE_value = PE_value*self.PE_correlated_mult
           
            if isinstance(PE_range, float): #Need this here because pars_range gets called later and is a 1D array.
                PE_range = PE_inputs
            #Check to make sure we do not have negative values
            #if PE_range[0] + PE_value < 0: CAN HAVE NEGATIVE PE VALUES!
                #pars_range[range_key_PE] = [0, PE_range[1] +PE_value, PE_range[2]]
            else:
                pars_range[range_key_PE] = [PE_range[0] +PE_value, PE_range[1] +PE_value, PE_range[2]]

            #print("Calculated range is " + str(pars_range[range_key][0]) + " " + str(pars_range[range_key][1]))
            pars_range['branching_ratio'] = pars_range['branching_ratios'][i]
            #pars_range['spinOrbitSplit'] = pars_range['spinOrbitSplit'][i]

            range_key_sos = 'spinOrbitSplitting'
            guess_key_sos = 'spinOrbitSplit'
            spinOrbitSplit_range = pars_range[range_key_sos]
            SOS1,SOS2 =spinOrbitSplit_range[0],spinOrbitSplit_range[1]
            sos = pars_range[guess_key_sos][i]

            pars_range[range_key_sos][0],pars_range[range_key_sos][1] = [spinOrbitSplit_range[0] + sos, spinOrbitSplit_range[1] + sos,]

           

            #New Sigma ranges for each peak
            range_key_sigma = 'Gaussian'
            guess_key_sigma = 'Sigma'

            sigma_inputs = np.array(sigma_inputs)
            if sigma_inputs.ndim > 1:
                sigma_range = sigma_inputs[self.sigma_correlated-1] #2D array
            else:
                sigma_range = sigma_inputs #1D array
           
            S1, S2 = sigma_range[0], sigma_range[1]    
           
            sigma_value = pars_range[guess_key_sigma][self.sigma_correlated-1]
            sigma_value = sigma_value*self.sigma_correlated_mult
           
            if isinstance(sigma_range, float): #Need this here because pars_range gets called later and is a 1D array.
                sigma_range = sigma_inputs

            #Check to make sure we do not have negative values
            if sigma_range[0] + sigma_value < 0:
                pars_range[range_key_sigma] = [0, sigma_range[1] +sigma_value, sigma_range[2]] 
            else:
                pars_range[range_key_sigma] = [sigma_range[0] +sigma_value, sigma_range[1] +sigma_value, sigma_range[2]] 
            
            #New Gamma ranges for each peak
            range_key_gamma = 'Lorentzian'
            guess_key_gamma = 'Gamma'

            gamma_inputs = np.array(gamma_inputs)
            if gamma_inputs.ndim > 1:
                gamma_range = gamma_inputs[self.gamma_correlated-1] #2D array
            else:
                gamma_range = gamma_inputs #1D array

           
            gamma_value = pars_range[guess_key_gamma][self.gamma_correlated-1]
            gamma_value = gamma_value*self.gamma_correlated_mult
            if isinstance(gamma_range, float): #Need this here because pars_range gets called later and is a 1D array.
                gamma_range = gamma_inputs
            #Check to make sure we do not have negative values
            if gamma_range[0] + gamma_value < 0:
                pars_range[range_key_gamma] = [0, gamma_range[1] +gamma_value, gamma_range[2]]
            else:
                pars_range[range_key_gamma] = [gamma_range[0] +gamma_value, gamma_range[1] +gamma_value, gamma_range[2]]

          
           



            '''
            #New user inputs of Sigma, Gamma, and Amplitude
            range_key_sigma = 'Gaussian'
            guess_key_sigma = 'Sigma'
            sigma_range = pars_range[range_key_sigma]
            S1,S2 =sigma_range[0],sigma_range[1]
            sigma_value = pars_range[guess_key_sigma][i]
            pars_range[range_key_sigma][0],pars_range[range_key_sigma][1] = [sigma_range[0] + sigma_value, sigma_range[1] + sigma_value,]

            range_key_gamma = 'Lorentzian'
            guess_key_gamma = 'Gamma'
            fwhm_range = pars_range[range_key_gamma]

            G1,G2 =fwhm_range[0],fwhm_range[1]
            gamma_value = pars_range[guess_key_gamma][i]
            pars_range[range_key_gamma][0],pars_range[range_key_gamma][1] = [fwhm_range[0] + gamma_value, fwhm_range[1] + gamma_value,]
            '''



            #Coster-Kronig Lorentz values:
            range_key_gamma_CK = 'Lorentzian Coster-Kronig'
            guess_key_gamma_CK = 'Gamma Coster-Kronig'
            fwhm_range_CK = pars_range[range_key_gamma_CK]
            
            CK1,CK2 =fwhm_range_CK[0],fwhm_range_CK[1]
            gamma_value_CK = pars_range[guess_key_gamma_CK][i]
            pars_range[range_key_gamma_CK][0],pars_range[range_key_gamma_CK][1] = [fwhm_range_CK[0] + gamma_value_CK, fwhm_range_CK[1] + gamma_value_CK]



            #New Amplitude ranges for each peak
            range_key_amp = 'Amplitude'
            guess_key_amp = 'Amp'

            amp_inputs = np.array(amp_inputs)
            if amp_inputs.ndim > 1:
                amp_range = amp_inputs[self.amp_correlated-1] #2D array
            else:
                amp_range = amp_inputs #1D array

            #A1,A2,A3 =amp_range[0],amp_range[1]
            amp_value = pars_range[guess_key_amp][self.amp_correlated-1]
            amp_value = amp_value*self.amp_correlated_mult
            
            if isinstance(amp_range, float): #Need this here because pars_range gets called later and is a 1D array.
                amp_range = amp_inputs
            #Check to make sure we do not have negative values
            if amp_range[0] + amp_value < 0:
                pars_range[range_key_amp] = [0, amp_range[1] +amp_value, amp_range[2]]
            else:
                pars_range[range_key_amp] = [amp_range[0] +amp_value, amp_range[1] +amp_value, amp_range[2]]
           
            


            self.peakArr[i] = peak(pars_range,peaks[i],is_singlet = is_singlet[i], is_coster_kronig = is_coster_kronig[i], PE_correlated = pars_range['PE_correlated'][i], PE_correlated_mult = pars_range['PE_correlated_mult'][i], sigma_correlated = pars_range['sigma_correlated'][i],sigma_correlated_mult = pars_range['sigma_correlated_mult'][i],gamma_correlated = pars_range['gamma_correlated'][i],gamma_correlated_mult = pars_range['gamma_correlated_mult'][i],amp_correlated = pars_range['amp_correlated'][i],amp_correlated_mult = pars_range['amp_correlated_mult'][i]) #ERROR HERE --> HOW TO MAKE XES_FIT ONLY SEE FIRST ARRAY IN AMPLITUDE???
            pars_range[range_key_sos][0],pars_range[range_key_sos][1] = SOS1,SOS2
        

        '''
        except:
            #Special option if youre creating a custom individual(i.e. for analysis)
            if pars_range =='':
                pass
            else:
                print("Error modding guesses")
                exit()
        '''

        #each index in the peaks/background array is the name of the peak/background type to be used
        k=0
        n=0
        '''
        for i in range(self.nBackgrounds):
            if backgrounds[i] == 'Shirley-Sherwood':
                self.bkgnArr = [None] * (self.nBackgrounds + self.nPeaks)
        '''

        

        for i in range(self.nBackgrounds):
       
            if backgrounds[i] == 'SVSC_shirley':
                self.nBackgrounds -=1
                pass
            elif backgrounds[i] == 'Shirley-Sherwood':

                #while n >= self.nPeaks:
                self.bkgnArr[k] = background(pars_range,backgrounds[i], peaks[n]) #Creating a unique shirley background for each peak...I think

                k +=1
                n +=1

            else:

                #n = 0 #Doesn't matter which peak we are looking at --> Only matters in Shirley case
                self.bkgnArr[k] = background(pars_range,backgrounds[i], peaks[n])
               
                k+=1

        
        #Moved down to account for calling ranges in background()
       
        for i in range(self.nPeaks): 
            
            if self.SVSC_toggle:
                self.peakArr[i].SVSC_toggle(self.SVSC_toggle) #activate peak shirley
          
            #pars_range[range_key][0],pars_range[range_key][1] = PE1,PE2
            for i in range(len(pars_range[range_key_PE])): pars_range[range_key_PE][i] = PE_range[i]
           
            pars_range[range_key_sos][0],pars_range[range_key_sos][1] = SOS1,SOS2

            #Added in sigma, gamma, and amplitude
            for i in range(len(pars_range[range_key_sigma])): pars_range[range_key_sigma][i] = sigma_range[i]
            for i in range(len(pars_range[range_key_gamma])): pars_range[range_key_gamma][i] = gamma_range[i]
            #pars_range[range_key_sigma][0],pars_range[range_key_sigma][1] = S1,S2
            #pars_range[range_key_gamma][0],pars_range[range_key_gamma][1] = G1,G2
            pars_range[range_key_gamma_CK][0],pars_range[range_key_gamma_CK][1] = CK1,CK2
            for i in range(len(pars_range[range_key_amp])): pars_range[range_key_amp][i] = amp_range[i]
            #pars_range[range_key_amp][0],pars_range[range_key_amp][1], pars_range[range_key_amp][2] = A1,A2,A3
           

        #Turn peak ranges back into 2D array to sort through all peaks otherwise it defaults to the last peaks parameter ranges
        pars_range[range_key_sigma] = full_sigma_range
        pars_range[range_key_gamma] = full_gamma_range
        pars_range[range_key_amp] = full_amp_range
        pars_range[range_key_PE] = full_energy_range

        

       

        '''
        for n in range(self.nPeaks):
            for i in range(self.nBackgrounds): #May have to iterate through peaks too if we want background to be peak dependent
                if backgrounds[i] == 'SVSC_shirley':
                    self.nBackgrounds -=1
                    pass
                else:

                    self.bkgnArr[k] = background(pars_range,backgrounds[i], peaks[n])
                    k+=1
        '''
         # Create dictionary of peaks and backgrounds
        for i in range(self.nPeaks):
            self.peakDict[f'peak_{i}'] = self.peakArr[i]
            

        for i in range(self.nBackgrounds):
            self.peakDict[f'bkgn_{i}'] = self.bkgnArr[i]

    def add_peak(self,peakType):
        
        self.peakArr.append(peak(self.pars_range,peakType))
    def add_bkgn(self,bkgnType):
        
        self.bkgnArr.append(background(self.pars_range,bkgnType))
    
    #adds all backgrounds and peaks as one y value array
    def getFit(self,x,y, backgrounds):
        #ISSUE WHEN ONE PEAK IS DL/GAUSS/DS --> MOVES RANGES OVER BY ONE PEAK TILL LAST PEAK WHICH IS CONSTANT. KICKS IN AT START OF SECOND GEN
        yFit = [0]*len(x)
        FWHM = [0]*self.nPeaks
        
       
        for i in range(self.nPeaks):
       
          
            '''
            asymD = self.peakArr[i].asymmetryDoniach
       

            gamma_correlated = int(self.peakArr[i].gamma_correlated)
            gamma_correlated_mult = self.peakArr[i].gamma_correlated_mult
            
            #THIS IS WRONG --> TAKING IN INOUT VALUES NOT THE VALUES FOUND IN XES_FIT
            self.PE_correlated = int(self.pars_range['PE_correlated'][i])
            self.PE_correlated_mult = self.pars_range['PE_correlated_mult'][i]
            self.sigma_correlated = int(self.pars_range['sigma_correlated'][i])
            self.sigma_correlated_mult = self.pars_range['sigma_correlated_mult'][i]
            self.gamma_correlated = int(self.pars_range['gamma_correlated'][i])
            self.gamma_correlated_mult = self.pars_range['gamma_correlated_mult'][i]
            
            self.amp_correlated = int(self.pars_range['amp_correlated'][i])
            self.amp_correlated_mult = self.pars_range['amp_correlated_mult'][i]

            
            PE_value = self.pars_range['PE'][self.PE_correlated-1] 
            PE_value = PE_value*self.PE_correlated_mult
            PE = PE_value

            gamma_value = self.pars_range['Gamma'][self.gamma_correlated-1]
            gamma_value = gamma_value*self.gamma_correlated_mult
            width = gamma_value

            sigma_value = self.pars_range['Sigma'][self.sigma_correlated-1]
            sigma_value = sigma_value*self.sigma_correlated_mult
            sigma = sigma_value
            

            amp_value = self.pars_range['Amp'][self.amp_correlated-1]
            amp_value = amp_value*self.amp_correlated_mult
            A = amp_value

            asym = self.peakArr[i].asymmetry
            singlet = self.peakArr[i].is_singlet
            coster_kronig = self.peakArr[i].is_coster_kronig
            if coster_kronig == True:
                width_CK = self.peakArr[i].lorentz_CK
            else:
                width_CK = self.peakArr[gamma_correlated-1].lorentz
            branch = self.peakArr[i].branching_ratio
            split = self.peakArr[i].spinOrbitSplit
            '''
            
     
            PE_correlated = int(self.peakArr[i].PE_correlated)
           
            PE_correlated_mult = self.peakArr[i].PE_correlated_mult
            sigma_correlated = int(self.peakArr[i].sigma_correlated)
            sigma_correlated_mult = self.peakArr[i].sigma_correlated_mult
            gamma_correlated = int(self.peakArr[i].gamma_correlated)
            gamma_correlated_mult = self.peakArr[i].gamma_correlated_mult
            amp_correlated = int(self.peakArr[i].amp_correlated)
            amp_correlated_mult = self.peakArr[i].amp_correlated_mult
            
            asymD = self.peakArr[i].asymmetryDoniach
            PE = self.peakArr[PE_correlated-1].photonEnergy
            PE = PE*PE_correlated_mult
            width = self.peakArr[gamma_correlated-1].lorentz
            width = width*gamma_correlated_mult
            sigma = self.peakArr[sigma_correlated-1].gaussian
            sigma = sigma*sigma_correlated_mult
            A = self.peakArr[amp_correlated-1].amp
            A = A*amp_correlated_mult
            asym = self.peakArr[i].asymmetry
            singlet = self.peakArr[i].is_singlet
            coster_kronig = self.peakArr[i].is_coster_kronig
            if coster_kronig == True:
                width_CK = self.peakArr[i].lorentz_CK
            else:
                width_CK = self.peakArr[gamma_correlated-1].lorentz
            branch = self.peakArr[i].branching_ratio
            split = self.peakArr[i].spinOrbitSplit

            #print("PEAK #:", i)
            if self.SVSC_toggle:
                peak_y,svsc_y = self.peakArr[i].getY(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)
                yFit += peak_y
                yFit += svsc_y
            else:
                yFit += self.peakArr[i].getY(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)
                FWHM += self.peakArr[i].getFWHM(x, PE, width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split)



        n = 0
        for i in range(self.nBackgrounds):
            PE_correlated = int(self.peakArr[n].PE_correlated)
            sigma_correlated = int(self.peakArr[n].sigma_correlated)
            gamma_correlated = int(self.peakArr[n].gamma_correlated)
            amp_correlated = int(self.peakArr[n].amp_correlated)
            


            asymD = self.peakArr[n].asymmetryDoniach
            PE = self.peakArr[PE_correlated-1].photonEnergy
            
            width = self.peakArr[gamma_correlated-1].lorentz
            sigma = self.peakArr[sigma_correlated-1].gaussian
            A = self.peakArr[amp_correlated-1].amp
            asym = self.peakArr[n].asymmetry
            singlet = self.peakArr[n].is_singlet
            coster_kronig = self.peakArr[n].is_coster_kronig
            branch = self.peakArr[n].branching_ratio
            split = self.peakArr[n].spinOrbitSplit
           
            yFit += self.bkgnArr[i].getY(x,y, PE , width, sigma, A, asym, asymD, singlet, coster_kronig, width_CK, branch, split,self.scale_var) #This is where backgrounds are actually calculated --> Is Shirley taking into account the multiple peaks??

            if backgrounds[i] == 'Shirley-Sherwood':
                n += 1
       
        return yFit
    def get(self):
        """
        Get the whole set
        """
        return (self.peakArr + self.bkgnArr)


    def get_params(self, for_output_file = False):
        params = []
        #fetches all the params as independent lists
       
        for i in range(len(self.peakArr)):
            params.append(self.peakArr[i].get(for_output_file)) #WHAT IS FOR_OUTPUT_FILE???
        for i in range(len(self.bkgnArr)):
            params.append(self.bkgnArr[i].get())

        #puts it in one array
        for i in range(1,len(params)):
            for k in range(len(params[i])):
                params[0].append(params[i][k])

        #print("Params : " + str(params[0]))
        return params[0]

    def get_peak(self,i):
        return self.peakArr[i].get()

    def get_peaks(self):
        return self.peakArr

    def get_background(self,i):
        return self.bkgnArr[i]
    def get_backgrounds(self):
        return self.bkgnArr

    def mutate_(self,chance):
        for peak in self.peakArr:
            peak.mutate(chance)
        for bkgn in self.bkgnArr:
            bkgn.mutate(chance)

    #forces a given peak to have the given values, returns 0 on success, -1 on failure
    def setPeak(self,i,param_arr):

        #param array comes in with its last element indicating its type
        peakType = param_arr[len(param_arr)-1]
        #if param_array is voigt, it comes in form [PE,Gauss,Lorentz,Amplitude,'Voigt']
        if peakType.lower() == 'voigt': #Idk if this is important or now???
            self.peakArr[i].set_voigt(param_arr)
            return 0
        elif peakType.lower() == 'double lorentzian':
            self.peakArr[i].set_doubleLorentz(param_arr)
            return 0
        elif peakType.lower() == 'gaussian':
            self.peakArr[i].set_gauss(param_arr)
            return 0
        elif peakType.lower() == 'lorentzian':
            self.peakArr[i].set_lorentz(param_arr)
            return 0
        elif peakType.lower() == 'doniach-sunjic':
            self.peakArr[i].set_doniachSunjic(param_arr)
            return 0
        else:
            
            return -1

    def setBkgn(self,i,param_arr):
        bkgnType = param_arr[len(param_arr)-1]
        #if param_array is voigt, it comes in form [PE,Gauss,Lorentz,Amplitude,'Voigt']
        if bkgnType.lower() == 'shirley-sherwood':
            self.bkgnArr[i].set_shirley_sherwood(param_arr)

    def verbose(self):
        """
        Print out the Populations
        """
        for i in range(self.npaths):
            self.Population[i].verbose()



import tkinter as tk
#from tkinter import Button, mainloop

from tkinter import ttk, Tk, N, W, E, S, StringVar, IntVar, DoubleVar, BooleanVar, Checkbutton, NORMAL, DISABLED, \
    scrolledtext, filedialog, messagebox, LabelFrame, Toplevel, END, TOP
from tkinter.font import Font
class App():
    def __init__(self):
        self.__version__ = 0.0
        self.root = Tk(className='XES Neo GUI')
        # use a lambda expression to pass arg to callback in button commands
        tk.Button(self.root, text='start', command=lambda: self.callback(True)).pack()
        tk.Button(self.root, text='stop', command=lambda: self.callback(False)).pack()
    def callback(self, flag):
        message = "message"
        for i in range(10):
            message += str(i)

        global loop
        if flag is True:
            print(message)
            # start the after method to call the callback every 100ms
            loop = self.root.after(100, self.callback, True)
        else:
            # cancel the after method if the flag is False
            self.root.after_cancel(loop)



    def Run(self):
        """
        Run the code
        """
        self.root.mainloop()

root = App()
root.Run()



import numpy as np

class xes_data():
    def __init__(self,fileName,skipLn, data_KE=False):
        self.fileName = fileName
        self.skipLn = skipLn #int represnting files to skip
        self.read_file_txt(fileName)
        
    #not yet complete i was just throwing down some starter code
    def read_file_txt(self,fileName):
        self.x,self.y = np.loadtxt(self.fileName,skiprows=self.skipLn,unpack=True)
        self.numP = len(self.x)

    
    def get_x(self,data_KE, data_XES):
        if data_KE == True:
            self.x = self.x[::-1]
        else:
            pass

        return self.x

    
    def get_y(self,scale_var):
        
        y_val = []
        #first = self.y[0]
        y_max = 0
        scale_val = 0

       
        first = 1

     
        y_max = self.y.max()
      
        #Scaling is done proportional to the maximum peak height. This also sets background to 0
        if y_max <= 0.001:
            scale_val = first*100000000
        elif y_max <= 0.01:
            scale_val = first*10000000
        elif y_max <= 0.1:
            scale_val = first*1000000
        elif y_max <= 1:
            scale_val = first*100000
        elif y_max <= 10:
            scale_val = first*10000
        elif y_max <= 100:
            scale_val = first*1000
        elif y_max <= 1000:
            scale_val = first*100
        elif y_max <= 10000:
            scale_val = first*10
        else:
            scale_val = first
      

        if scale_var == True:
            #y_val = self.y/first #Dividing every element by the first value 
            
            y_val = self.y*scale_val #Multiply by 1000 to scale it
           
        else:
            y_val = self.y
       
        
        return y_val
        


# Import Library
import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from xes_neo.input_arg import *
from xes_neo.helper import timecall

if timeing_mode:
# %matplotlib inline
    t1 = timecall()

from psutil import cpu_count
# Set the number of threads
import os
import sys
from operator import itemgetter
import numpy as np
import pathlib
import copy
import logging
import operator
import datetime
import random
import csv


import matplotlib as mpl
import matplotlib.pyplot as plt


if timeing_mode:
    initial_elapsed = timecall()- t1
    print('Inital import function took %.2f second' %initial_elapsed)



# global os
__version__ = '0.0.16'
__author__ = 'Jeff Terry, Alaina Humiston, Miu Lun Lau, Min Long'
__email__ = 'terryj@iit.edu, thompson9@hawk.iit.edu, andylau@u.boisestate.edu, minlong@boisestate.edu'


# Note
# variables only get save in __init__.py, but not everywhere else.
# variable does get save across, this show be the final location for running the script

#



#File of all literature information into the BE, sigma/gamma, and spin-orbitting splitting ranges for specific elemental compositions
#Could also just move all this information into the periodic table file
class ElementalComposition:
    def __init__(self,element,photoelectronLine, BE_guess):
        elementalComps = ['C', 'SiC', 'CO2'] #This is for testing
        self.elementalComps = elementalComps #--> Make input array for Fitting parameters tab that appends this info
        #Data Presented as C, SiC, CO2 etc.
        self.element = element
        self.photoelectronLine = photoelectronLine
        self.BE_lit, self.is_singlet, self.so_split, self.branching_ratio = self.getParams(self.element,photoelectronLine)
        self.BE_guess = BE_guess

    def getParamRanges(self):

        if self.element == 'C': #Should I make seperate function calls for each element or just have a lot of if statements?

            for i in self.elementalComps:
                if i == 'C':
                    BE_guess[i] = 284.5 #We want to append these values to an array for each peak. Call BE_guess from xps.py?
                elif i == 'SiC':
                    BE_guess[i] = 282.0



# coding: utf-8
from __future__ import print_function, unicode_literals
import sys
import codecs
from setuptools import setup, find_packages
from xes_neo._version import __version__, __author__, __email__



def long_description():
    with codecs.open('README.md', 'rb') as readme:
        if not sys.version_info < (3, 0, 0):
            return readme.read().decode('utf-8')


setup(
    name='xes_neo',
    version=__version__,
    packages=find_packages(),

    author=__author__,
    author_email=__email__,
    keywords=['GA', 'XES','analysis'],
    description='XES Analysis using GA',
    long_description=long_description(),
    url='https://github.com/lanl/XES_Neo_Public.git',
    download_url='https://github.com/lanl/XES_Neo_Public/tarball/main',
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'numpy',
        'numba',
        'attrs',
        'matplotlib',
        'scipy',
        'psutil'
    ],
    entry_points={
        'console_scripts': [
            'xes_neo = xes_neo.xes:main',
            'xes_neo_gui = xes_neo.gui.xes_neo_gui:main',
        ]
    },
    license='GPLv3',
)



from multiprocessing import Process, Queue
import os
import signal
import subprocess
from multiprocessing import Pool
from tkinter import ttk, Tk, N, W, E, S, StringVar, IntVar, DoubleVar, BooleanVar, Checkbutton, NORMAL, DISABLED, \
    scrolledtext, filedialog, messagebox, LabelFrame, Toplevel, END, TOP

#proc_lst = []

def F(x):
    #global proc_lst
    proc = subprocess.Popen(x, shell=True, preexec_fn=os.setsid)
    #proc_lst.append(proc.pid)
    #proc_lst.append(os.getpgid(proc.pid))
    #print(proc_lst)
    print("PROC ID: ", proc.pid)
    print(type(proc.pid))
    # proc.wait()
    return proc.pid

class SeriesInstance():
    def __init__(self):
        self.root = Tk()
        #self.numbers = Queue()
        #for i in range(4):
         #   self.numbers.put('exec nano_neo -i test/test_nanoindent_w_Params.ini')
        self.numbers = ['exec xes_neo -i test/test_xes_w_Params.ini', 'exec xes_neo -i test/test_xes_w_Params.ini', 'exec xes_neo -i test/test_xes_w_Params.ini']
        self.F = F
        self.proc_lst = []
        ttk.Button(self.root, text='start', command=self.start).pack()
        ttk.Button(self.root, text='stop', command=self.stop).pack()
        #semaphore = multiprocessing.Semaphore(1)
        self.p = Pool(maxtasksperchild=1)
        #self.p = multiprocessing.Semaphore(1)

    def run(self):
        self.root.mainloop()

    def start(self):
        #while not self.numbers.empty():
        #p = Process(target=F, args=(self.numbers,))
        #p.start()
        #self.proc_lst.append(p)
        #p.join()
        #p.close()
        self.proc_lst.append(self.p.map(self.F, self.numbers))
        print("Proc list inside start: ", self.proc_lst)
        self.p.join()
        self.p.close()
        #return self.proc_lst

    def stop(self):
        print(self.proc_lst)
        print("IN stop")
        print("IN STOP")
        print("IN stop")
        print("IN STOP")
        print("IN stop")
        print("IN STOP")
        print("IN stop")
        print("IN STOP")
        print("IN stop")
        print("IN STOP")
        print("IN stop")
        print("IN STOP")
        for each in self.proc_lst[0]:
            print("KILLING: ", each)
            os.kill(each, signal.SIGTERM)
        if hasattr(self, 'proc'):
            self.proc.kill()
        print("Attempting to kill pool:")
        self.p.close()
        self.p.terminate()
        self.p.join()

root = SeriesInstance()
root.run()



import numpy as np

class xes_data():
    def __init__(self,fileName,skipLn):
        self.fileName = fileName
        self.skipLn = skipLn #int represnting files to skip
       
        self.read_file_txt()

    #not yet complete i was just throwing down some starter code
    def read_file_txt(self):
        self.x,self.y = np.loadtxt(self.fileName,skiprows=self.skipLn,unpack=True)
        self.numP = len(self.x)
       
        

    def get_x(self):
        return self.x

    def get_y(self, scale_var):
        y_val = []
        first = 1
        y_max = self.y.max()

        #Scaling is done proportional to the maximum peak height. This also sets background to 0
        if y_max <= 0.001:
            scale_val = first*100000000
        elif y_max <= 0.01:
            scale_val = first*10000000
        elif y_max <= 0.1:
            scale_val = first*1000000
        elif y_max <= 1:
            scale_val = first*100000
        elif y_max <= 10:
            scale_val = first*10000
        elif y_max <= 100:
            scale_val = first*1000
        elif y_max <= 1000:
            scale_val = first*100
        elif y_max <= 10000:
            scale_val = first*10
        else:
            scale_val = first
            
        if scale_var == True:
            
            #y_val = self.y/first #Dividing every element by the first value 
            
            y_val = self.y*scale_val #Multiply by 1000 to scale it
          
        else:
            y_val = self.y
        
        min_y = np.min(self.y)
        
        '''
        for i in range(len(self.y)):
            self.y[i] = self.y[i] - min_y #setting background to zero. Not sure if this is applied everywhere (being seen) right now 
        y_val = self.y
        '''
        
       

        return y_val
    
    #Subtracting off last y value from data (makes the data go to zero to help with baseline range)
    '''
    def get_y(self):
        new_y = [0]*len(self.y)
        for i in range(len(self.y)):
            new_y[i] = self.y[i] - self.y[-1]
        print(new_y)
        #return self.y
        return new_y
    '''




import tkinter as tk

import numpy as np
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.figure import Figure
#import NanoIndent_Analysis
import xes_neo.gui.xes_analysis2

class Analysis_plot:
    def __init__(self, frame, data_KE = False, data_XES = False):
        
        self.frame = frame
        self.fig = Figure(figsize=(6.4,4.8), dpi=100)
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.frame)
        # Create initial figure canvas
        self.canvas.get_tk_widget().grid(column=2, row=1, rowspan=13, columnspan=5, sticky="nsew",
                                         padx=5, pady=5)
        self.ax = self.fig.add_subplot(111)
        # create toolbar
        self.toolbarFrame = tk.Frame(master=self.frame)
        self.toolbarFrame.grid(column=2, row=0, rowspan=1, columnspan=5, sticky="nsew")
        toolbar = NavigationToolbar2Tk(self.canvas, self.toolbarFrame)
        self.params = {}

    def initial_parameters(self,dir,params,scale_var,data_KE,data_XES,peakType, title):
        dir = str(dir.get())
        self.xes_analysis = xes_neo.gui.xes_analysis2.xes_analysis(dir,params, peakType)
        self.xes_analysis.extract_data(scale_var,plot_err=False) #Maybe add scale_var to this function to make it seen in xes_individual?
        self.xes_analysis.score()
        self.totalArea, self.areas = self.xes_analysis.analyze()
        self.xes_analysis.analyze()
        self.fig.clf()
        self.xes_analysis.plot_data(scale_var, data_KE, data_XES, title=title,fig_gui = self.fig)
        self.canvas.draw()
        return self.xes_analysis.get_params()
        '''
        self.nano_analysis.score()
        self.nano_analysis.calculate_parameters(verbose=False)
        self.fig.clf()
        self.nano_analysis.plot_data(title=title,fig_gui=self.fig)
        self.canvas.draw()
        return self.nano_analysis.get_params()
        '''
class Data_plot:
    def __init__(self, frame, data_KE = False, data_XES = False):
        self.frame = frame
        self.fig = Figure(figsize=(6.4, 4.8), dpi=100)
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.frame)
        # Create initial figure canvas
        self.canvas.get_tk_widget().grid(column=0, row=1, rowspan=4, columnspan=8, sticky="nsew",
                                         padx=5, pady=5)
        self.ax = self.fig.add_subplot(111)
        # create toolbar
        self.toolbarFrame = tk.Frame(master=self.frame)
        self.toolbarFrame.grid(column=3, row=0, rowspan=1, columnspan=5, sticky="w")
        toolbar = NavigationToolbar2Tk(self.canvas, self.toolbarFrame)
        self.params = {}

    def initial_parameters(self,data_obj, data_KE, data_XES):
        """
        fileraw
        """
        self.data_obj = data_obj
        self.title = 0
        
        if data_KE == False:
            self.xlabel = 'Binding Energy (eV)'
     
        else:
            self.xlabel = 'Kinetic Energy (eV)'

        if data_XES == True:
            self.xlabel = 'Photon Energy (eV)'
       
        self.ylabel= 'Intensity (a.u.)'
    #def plot written by evan 
    def plot(self,x_data_array,y_data_array, param_label,title, scale_var,data_KE, data_XES):
        x = self.data_obj.get_x(data_KE, data_XES)
        y = self.data_obj.get_y(scale_var)

        min_y = np.min(y)
       
        for i in range(len(y)):
            y[i] = y[i] - min_y
         
        self.ax.clear()
        '''
        for i in range(len(x_data_array)):
            
            self.ax.plot(x_data_array[i],y_data_array[i], 'b.', label = param_label)
        '''
        #global data_KE
        self.ax.plot(x,y,color='gray', linestyle='None', marker='o', markeredgewidth='0.3', markerfacecolor='None', markeredgecolor='gray', markersize='5', label='Data')
        
        if data_KE == False:
            self.ax.invert_xaxis()   
        else:
            pass
        if data_XES == True:
            self.ax.invert_xaxis()  
        

        #self.ax.invert_xaxis()
        self.ax.legend()
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)
        self.ax.set_title(title)
        self.fig.tight_layout()

        self.canvas.draw()

    